'''
gff2fasta.py - output sequences from genomic features
=====================================================

:Tags: Genomics Intervals Sequences GFF Fasta Transformation

Purpose
-------

This script outputs the genomic sequences for intervals within
a :term:`gff` or :term: `gtf` formatted file.

The ouput can be optionally masked and filtered.

Usage
-----

If you want to convert a ``features.gff`` file with intervals information
into a :term:`fasta` file containing the sequence of each interval, use this
script as follows::

   python gff2fasta.py --genome-file=hg19 < features.gff > features.fasta

The input can also be a :term:`gtf` formatted file. In that case, use the
``--is-gtf`` option::

   python gff2fasta.py --genome-file=hg19 --is-gtf < features.gtf >\
 features.fasta

If you want to add a polyA tail onto each transcript you can use the `extend`
options:

   python gff2fasta.py --genome-file=hg19 --is-gtf
   --extend-at=3 --extend-by=125 --extend-with=A
   < features.gtf > features.fasta

If you want to merge the sequence of similar features together, please use
``--merge-overlapping``::

   python gff2fasta.py --genome-file=hg19 --merge-overlapping < features.gff >\
 features.fasta

It is possible to filter the output by selecting a minimum or maximum number
of nucleotides in the resultant fasta sequence with ``--max-length`` or
``--min-interval-length`` respectively::

   python gff2fasta.py --genome-file=hg19 --max-length=100\
 < features.gff > features.fasta

Or you can also filter the output by features name with the ``--feature``
option::

   python gff2fasta.py --genome-file=hg19 --feature=exon < features.gff\
 > features.fasta

On the other hand, low-complexity regions can be masked with the ``--masker``
option and a given :term:`gff` formatted file::

   python gff2fasta.py --genome-file=hg19 --masker=dust\
 --maskregions-bed-file=intervals.gff < features.gff > features.fasta

where ``--masker`` can take the following values: ``dust``, ``dustmasker``,
and ``softmask``.

Options
-------

``--is-gtf``
  Tells the script to expect a :term:`gtf` format file

``--genome-file``
  PATH to Fasta file of genome build to use

``--merge-overlapping``
  Merge features in :term:`gtf`/:term:`gff` file that are adjacent and share
  attributes

``--method=filter --filter-method``
  Filter on a :term:`gff` feature such as ``exon`` or ``CDS``

``--maskregions-bed-file``
  Mask sequences in intervals in :term:`gff` file

``--remove-masked-regions``
  Remove sequences in intervals in :term:`gff` file rather than masking them

``--min-interval-length``
  Minimum output sequence length

``--max-length``
  Maximum output sequence length

``--extend-at``
  Extend sequence at 3', 5' or both end.  Optionally '3only' or '5only' will
  return only the 3' or 5' extended sequence

``--extend-by``
  Used in conjunction with ``--extend-at``, the number of nucleotides to extend
  by

``--extend-with``
  Optional. Used in conjunction with ``--extend-at`` and ``--extend-by``.
  Instead of extending by the genomic sequence, extend by this string repeated
  n times, where n is --entend-by


``--masker``
  Masker type to use: dust, dustmasker, soft or none

``--fold-at``
  Fold the fasta sequence every n bases

``--naming-attribute``
  Use this attribute to name the fasta entries

Command line options
--------------------
'''

import sys
import CGAT.Experiment as E
import CGAT.GTF as GTF
import CGAT.Genomics as Genomics
import CGAT.IOTools as IOTools
import CGAT.IndexedFasta as IndexedFasta
import CGAT.Intervals as Intervals
import CGAT.Masker as Masker
import bx.intervals.intersection


def main(argv=None):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if argv is None:
        argv = sys.argv

    parser = E.OptionParser(
        version="%prog version: $Id$",
        usage=globals()["__doc__"])

    parser.add_option("--is-gtf", dest="is_gtf", action="store_true",
                      help="input is gtf instead of gff.")

    parser.add_option("-g", "--genome-file", dest="genome_file", type="string",
                      help="filename with genome [default=%default].")

    parser.add_option(
        "-m", "--merge-adjacent", dest="merge", action="store_true",
        help="merge adjacent intervals with the same attributes."
        " [default=%default]")

    parser.add_option(
        "-e", "--feature", dest="feature", type="string",
        help="filter by a feature, for example 'exon', 'CDS'."
        " If set to the empty string, all entries are output "
        "[%default].")

    parser.add_option(
        "-f", "--maskregions-bed-file", dest="filename_masks",
        type="string", metavar="gff",
        help="mask sequences with regions given in gff file "
        "[%default].")

    parser.add_option(
        "--remove-masked-regions", dest="remove_masked_regions",
        action="store_true",
        help="remove regions instead of masking [%default].")

    parser.add_option(
        "--min-interval-length", dest="min_length", type="int",
        help="set minimum length for sequences output "
        "[%default]")

    parser.add_option(
        "--max-length", dest="max_length", type="int",
        help="set maximum length for sequences output "
        "[%default]")

    parser.add_option(
        "--extend-at", dest="extend_at", type="choice",
        choices=("none", "3", "5", "both", "3only", "5only"),
        help="extend at no end, 3', 5' or both ends. If "
        "3only or 5only are set, only the added sequence "
        "is returned [default=%default]")

    parser.add_option(
        "--header-attributes", dest="header_attr",
        action="store_true",
        help="add GFF entry attributes to the FASTA record"
        " header section")

    parser.add_option(
        "--extend-by", dest="extend_by", type="int",
        help="extend by # bases [default=%default]")

    parser.add_option(
        "--extend-with", dest="extend_with", type="string",
        help="extend using base [default=%default]")

    parser.add_option(
        "--masker", dest="masker", type="choice",
        choices=("dust", "dustmasker", "softmask", "none"),
        help="apply masker [%default].")

    parser.add_option(
        "--fold-at", dest="fold_at", type="int",
        help="fold sequence every n bases[%default].")

    parser.add_option(
        "--fasta-name-attribute", dest="naming_attribute", type="string",
        help="use attribute to name fasta entry. Currently only compatable"
        " with gff format [%default].")

    parser.set_defaults(
        is_gtf=False,
        genome_file=None,
        merge=False,
        feature=None,
        filename_masks=None,
        remove_masked_regions=False,
        min_length=0,
        max_length=0,
        extend_at=None,
        extend_by=100,
        extend_with=None,
        masker=None,
        fold_at=None,
        naming_attribute=False,
        header_attr=False,
    )

    (options, args) = E.Start(parser)

    if options.genome_file:
        fasta = IndexedFasta.IndexedFasta(options.genome_file)
        contigs = fasta.getContigSizes()

    if options.is_gtf:
        iterator = GTF.transcript_iterator(GTF.iterator(options.stdin))
    else:
        gffs = GTF.iterator(options.stdin)
        if options.merge:
            iterator = GTF.joined_iterator(gffs)
        else:
            iterator = GTF.chunk_iterator(gffs)

    masks = None
    if options.filename_masks:
        masks = {}
        with IOTools.openFile(options.filename_masks, "r") as infile:
            e = GTF.readAsIntervals(GTF.iterator(infile))

        # convert intervals to intersectors
        for contig in list(e.keys()):
            intersector = bx.intervals.intersection.Intersecter()
            for start, end in e[contig]:
                intersector.add_interval(bx.intervals.Interval(start, end))
            masks[contig] = intersector

    ninput, noutput, nmasked, nskipped_masked = 0, 0, 0, 0
    nskipped_length = 0
    nskipped_noexons = 0

    feature = options.feature

    # iterator is a list containing groups (lists) of features.
    # Each group of features have in common the same transcript ID, in case of
    # GTF files.
    for ichunk in iterator:

        ninput += 1

        if feature:
            chunk = [x for x in ichunk if x.feature == feature]
        else:
            chunk = ichunk

        if len(chunk) == 0:
            nskipped_noexons += 1
            E.info("no features in entry from "
                   "%s:%i..%i - %s" % (ichunk[0].contig,
                                       ichunk[0].start,
                                       ichunk[0].end,
                                       str(ichunk[0])))
            continue

        contig, strand = chunk[0].contig, chunk[0].strand
        
        if options.is_gtf:
            name = chunk[0].transcript_id
        else:
            if options.naming_attribute:
                attr_dict = {x.split("=")[0]: x.split("=")[1]
                             for x in chunk[0].attributes.split(";")}
                name = attr_dict[options.naming_attribute]
            else:
                name = str(chunk[0].attributes)

        lcontig = contigs[contig]
        positive = Genomics.IsPositiveStrand(strand)
        intervals = [(x.start, x.end) for x in chunk]
        intervals.sort()

        if masks:
            if contig in masks:
                masked_regions = []
                for start, end in intervals:
                    masked_regions += [(x.start, x.end)
                                       for x in masks[contig].find(start, end)]

                masked_regions = Intervals.combine(masked_regions)
                if len(masked_regions):
                    nmasked += 1

                if options.remove_masked_regions:
                    intervals = Intervals.truncate(intervals, masked_regions)
                else:
                    raise NotImplementedError("unimplemented")

                if len(intervals) == 0:
                    nskipped_masked += 1
                    if options.loglevel >= 1:
                        options.stdlog.write("# skipped because fully masked: "
                                             "%s: regions=%s masks=%s\n" %
                                             (name,
                                              str([(x.start,
                                                    x.end) for x in chunk]),
                                              masked_regions))
                    continue

        out = intervals

        if options.extend_at and not options.extend_with:
            if options.extend_at == "5only":
                intervals = [(max(0, intervals[0][0] - options.extend_by),
                              intervals[0][0])]
            elif options.extend_at == "3only":
                intervals = [(intervals[-1][1],
                              min(lcontig,
                                  intervals[-1][1] + options.extend_by))]
            else:
                if options.extend_at in ("5", "both"):
                    intervals[0] = (max(0,
                                        intervals[0][0] - options.extend_by),
                                    intervals[0][1])
                if options.extend_at in ("3", "both"):
                    intervals[-1] = (intervals[-1][0],
                                     min(lcontig,
                                         intervals[-1][1] + options.extend_by))

        if not positive:
            intervals = [(lcontig - x[1], lcontig - x[0])
                         for x in intervals[::-1]]
            out.reverse()

        s = [fasta.getSequence(contig, strand, start, end)
             for start, end in intervals]
        # IMS: allow for masking of sequences
        s = Masker.maskSequences(s, options.masker)
        l = sum([len(x) for x in s])
        if (l < options.min_length or
                (options.max_length and l > options.max_length)):
            nskipped_length += 1
            if options.loglevel >= 1:
                options.stdlog.write("# skipped because length out of bounds "
                                     "%s: regions=%s len=%i\n" %
                                     (name, str(intervals), l))
                continue

        if options.extend_at and options.extend_with:
            extension = "".join((options.extend_with,) * options.extend_by)

            if options.extend_at in ("5", "both"):
                s[1] = extension + s[1]
            if options.extend_at in ("3", "both"):
                s[-1] = s[-1] + extension

        if options.fold_at:
            n = options.fold_at
            s = "".join(s)
            seq = "\n".join([s[i:i+n] for i in range(0, len(s), n)])
        else:
            seq = "\n".join(s)

        if options.header_attr:
            attributes = " ".join([":".join([ax, ay]) for ax, ay in chunk[0].asDict().items()])
            options.stdout.write(">%s %s:%s:%s feature:%s %s\n%s\n" % (name,
                                                                       contig,
                                                                       strand,
                                                                       ";".join(
                                                                           ["%i-%i" %
                                                                            x for x in out]),
                                                                       chunk[0].feature,
                                                                       attributes,
                                                                       seq))
        else:
            options.stdout.write(">%s %s:%s:%s\n%s\n" % (name,
                                                         contig,
                                                         strand,
                                                         ";".join(
                                                             ["%i-%i" %
                                                              x for x in out]),
                                                         seq))

        noutput += 1

    E.info("ninput=%i, noutput=%i, nmasked=%i, nskipped_noexons=%i, "
           "nskipped_masked=%i, nskipped_length=%i" %
           (ninput, noutput, nmasked, nskipped_noexons,
            nskipped_masked, nskipped_length))

    E.Stop()

if __name__ == "__main__":
    sys.exit(main(sys.argv))



'''IndexedGenome.py - Random access to interval lists
==================================================

This module provides a consistent front-end to various interval containers.

Two implementations are available:

NCL
   Nested containment lists as described in
   http://bioinformatics.oxfordjournals.org/content/23/11/1386.short. The
   implemenation was taken from `pygr
   <http://code.google.com/p/pygr>`_.

quicksect
   Quicksect algorithm used in Galaxy, see `here
   <https://github.com/brentp/quicksect>`_.  This requires python.bx
   to be installed. The benefit of quicksect is that it allows also
   quick retrieval of intervals that are closest before or after an query.

The principal clas is :class:`IndexedGenome` which uses NCL and stores
a value associated with each interval. :class:`Quicksect` is equivalent
to :class:`IndexedGenome` but uses quicksect. The :class:`Simple` is a
light-weight version of :class:`IndexedGenome` that does not store a
value and thus preserves space.

The basic usage is::

   from IndexedGenome import IndexedGenome
   index = IndexedGenome()
   for contig, start, end, value in intervals:
      index.add(contig, start, end, value)

   print index.contains("chr1", 1000, 2000)
   print index.get("chr1", 10000, 20000)

The index is built in memory.

Reference
---------

'''
from CGAT import NCL as ncl
from bx.intervals.intersection import Intersecter, Interval


class IndexedGenome:

    '''Genome with indexed intervals.
    '''

    index_factory = ncl.NCL

    def __init__(self):
        self.mIndex = {}

    def add(self, contig, start, end, value):

        if contig not in self.mIndex:
            self.mIndex[contig] = self.index_factory()
        self.mIndex[contig].add(start, end, value)

    def __getitem__(self, args):
        '''return intervals overlapping with key.'''
        if args[0] not in self.mIndex:
            raise KeyError("contig %s not in index" % args[0])

        return self.mIndex[args[0]].find(args[1], args[2])

    def contains(self, contig, start, end):
        if contig not in self.mIndex:
            return False
        return len(list(self.mIndex[contig].find(start, end))) > 0

    def get(self, contig, start, end):
        '''return intervals overlapping with key.'''
        if contig not in self.mIndex:
            raise KeyError("contig %s not in index" % contig)

        return self.mIndex[contig].find(start, end)

    def __len__(self):
        '''return number of contigs.'''
        return len(self.mIndex)


class Simple(IndexedGenome):

    '''index intervals without storing a value.'''
    index_factory = ncl.NCLSimple

    def __init__(self, *args, **kwargs):
        IndexedGenome.__init__(self, *args, **kwargs)

    def add(self, contig, start, end):

        if contig not in self.mIndex:
            self.mIndex[contig] = self.index_factory()
        self.mIndex[contig].add(start, end)


class Quicksect(IndexedGenome):

    '''index intervals using quicksect.

    Permits finding closest interval in case there is
    no overlap.
    '''
    index_factory = Intersecter

    def __init__(self, *args, **kwargs):
        IndexedGenome.__init__(self, *args, **kwargs)

    def add(self, contig, start, end, value):

        if contig not in self.mIndex:
            self.mIndex[contig] = self.index_factory()
        self.mIndex[contig].add_interval(Interval(start, end, value))

    def get(self, contig, start, end):
        '''return intervals overlapping with key.'''
        if contig not in self.mIndex:
            raise KeyError("contig %s not in index" % contig)

        return [(x.start, x.end, x.value)
                for x in self.mIndex[contig].find(start, end)]

    def before(self, contig, start, end, num_intervals=1, max_dist=2500):
        '''get closest interval before *start*.'''
        if contig not in self.mIndex:
            raise KeyError("contig %s not in index" % contig)
        return [(x.start, x.end, x.value)
                for x in self.mIndex[contig].before_interval(
                    Interval(start, end),
                    num_intervals=1,
                    max_dist=max_dist)]

    def after(self, contig, start, end, num_intervals=1, max_dist=2500):
        '''get closest interval after *end*.'''
        if contig not in self.mIndex:
            raise KeyError("contig %s not in index" % contig)
        return [(x.start, x.end, x.value)
                for x in self.mIndex[contig].after_interval(
                    Interval(start, end),
                    num_intervals=1,
                    max_dist=max_dist)]



'''
gff2psl.py - convert from gff to psl
====================================

:Tags: Genomics Intervals GFF PSL Conversion

Purpose
-------

This scripts converts from a :term:`gff` formatted
file to a :term:`psl` formatted file.
The output can be modified by the following command line options:

--allow-duplicates
    keep duplicate entries from gff/gtf input file

--genome-file
    restrict output to gff/gtf entries with contigs in fasta file

--queries-tsv-file
    restrict output to queries in fasta file

Usage
-----

Example::

   python gff2psl.py < in.gff > out.psl

Type::

   python gff2psl.py --help

for command line help.
genome-file

Command line options
--------------------
'''

import sys
import CGAT.Experiment as E
import CGAT.IndexedFasta as IndexedFasta
import CGAT.Blat as Blat
import CGAT.GTF as GTF
import alignlib_lite
import CGAT.Intervals as Intervals


def main(argv=None):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if argv is None:
        argv = sys.argv

    parser = E.OptionParser(
        version="%prog version: $Id: gff2psl.py 2781 2009-09-10 11:33:14Z andreas $", usage=globals()["__doc__"])

    parser.add_option("--is-gtf", dest="is_gtf", action="store_true",
                      help="input is gtf.")

    parser.add_option("--no-header", dest="with_header", action="store_false",
                      help="do not output BLAT header [default=%default].")

    parser.add_option("-g", "--genome-file", dest="genome_file", type="string",
                      help="filename with genome.")

    parser.add_option("--queries-tsv-file", dest="input_filename_queries", type="string",
                      help="fasta filename with queries [default=%default].")

    parser.add_option("--allow-duplicates", dest="allow_duplicates", action="store_true",
                      help="""permit duplicate entries. Adjacent exons of a transcript will still be merged [default=%default]."""  )

    parser.set_defaults(is_gtf=False,
                        genome_file=None,
                        with_header=True,
                        allow_duplicates=False,
                        test=None)

    (options, args) = E.Start(parser, add_pipe_options=True)

    if options.genome_file:
        genome_fasta = IndexedFasta.IndexedFasta(options.genome_file)
    else:
        genome_fasta = None

    if options.input_filename_queries:
        queries_fasta = IndexedFasta.IndexedFasta(
            options.input_filename_queries)
    else:
        queries_fasta = None

    ninput, noutput, nskipped = 0, 0, 0

    if options.is_gtf:
        iterator = GTF.transcript_iterator(GTF.iterator_filtered(GTF.iterator(sys.stdin),
                                                                 feature="exon"),
                                           strict=not options.allow_duplicates)
    else:
        iterator = GTF.joined_iterator(GTF.iterator(sys.stdin))

    if options.with_header:
        options.stdout.write(Blat.Match().getHeader() + "\n")

    for gffs in iterator:

        if options.test and ninput >= options.test:
            break

        ninput += 1

        result = alignlib_lite.py_makeAlignmentBlocks()

        xstart = 0

        intervals = Intervals.combine([(gff.start, gff.end) for gff in gffs])

        for start, end in intervals:
            xend = xstart + end - start

            result.addDiagonal(xstart, xend,
                               start - xstart)
            xstart = xend

        entry = Blat.Match()
        entry.mQueryId = gffs[0].transcript_id
        entry.mSbjctId = gffs[0].contig
        entry.strand = gffs[0].strand

        if genome_fasta:
            if entry.mSbjctId in genome_fasta:
                entry.mSbjctLength = genome_fasta.getLength(entry.mSbjctId)
            else:
                entry.mSbjctLength = result.getColTo()

        if queries_fasta:
            if entry.mQueryId in queries_fasta:
                entry.mQueryLength = queries_fasta.getLength(entry.mQueryId)
        else:
            entry.mQueryLength = result.getRowTo()

        entry.fromMap(result)

        options.stdout.write(str(entry) + "\n")
        noutput += 1

    E.info("ninput=%i, noutput=%i, nskipped=%i" % (ninput, noutput, nskipped))

    E.Stop()

if __name__ == "__main__":
    sys.exit(main(sys.argv))



'''Blat.py - tools for working with PSL formatted files and data
=============================================================

This module provides a class to parse :term:`PSL` formatted
files such as those output by the BLAT tool.

This module defines the :class:`Blat.Match` class representing a
single entry and a series of iterators to iterate of :term:`PSL`
formatted files (:func:`iterator`, :func:`iterator_target_overlap`,
...).

Reference
---------

'''
import copy
import string
import collections

try:
    import alignlib_lite
except ImportError:
    pass

from CGAT import Components as Components
from CGAT import Experiment as E


class Error(Exception):
    """Base class for exceptions in this module."""

    def __str__(self):
        return str(self._message)

    def _get_message(self):
        return self._message

    def _set_message(self, message):
        self._message = message
    message = property(_get_message, _set_message)


class ParsingError(Error):

    """Exception raised for errors while parsing

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message, line=None):
        if line:
            self.message = message + " at line:" + line
        else:
            self.message = message

HEADER = """psLayout version 3

match   mis-    rep.    N's     Q gap   Q gap   T gap   T gap   strand  Q       Q       Q       Q       T       T       T       T       block   blockSizes      qStarts  tStarts
        match                   count   bases   count   bases           name    size    start   end     name    size    start   end     count
---------------------------------------------------------------------------------------------------------------------------------------------------------------"""


class Match:

    """a :term:`psl` formatted alignment.

    Block coordinates are on the forward strand for target and on
    the forward/reverse strand for the query depending on the strand.

    The fields mQueryFrom/To and mSbjctFrom/To are always on the forward
    strand.
    """

    def __init__(self):
        self.mNMatches = 0
        self.mNMismatches = 0
        self.mNRepMatches = 0
        self.mNns = 0
        self.mQueryNGapsCounts = 0
        self.mQueryNGapsBases = 0
        self.mSbjctNGapsCounts = 0
        self.mSbjctNGapsBases = 0
        self.strand = "+"
        self.mQueryId = ""
        self.mQueryLength = 0
        self.mQueryFrom = 0
        self.mQueryTo = 0
        self.mSbjctId = ""
        self.mSbjctLength = 0
        self.mSbjctFrom = 0
        self.mSbjctTo = 0
        self.mNBlocks = 0
        self.mBlockSizes = []
        self.mQueryBlockStarts = []
        self.mSbjctBlockStarts = []

    def convertCoordinates(self):
        """convert coordinates.

        This rescales the block positions so that they start at 0 and converts
        the query to forward and the sbjct to forward/reverse coordinates.

        About the psl psl format from the manual at
        http://genome.ucsc.edu/google/goldenPath/help/pslSpec.html

        ::
           In general the coordinates in psl files are "zero based
           half open." The first base in a sequence is numbered zero
           rather than one.  When representing a range the end
           coordinate is not included in the range. Thus the first 100
           bases of a sequence are represented as 0-100, and the
           second 100 bases are represented as 100-200.

           There is a another little unusual feature in the .psl
           format. It has to do with how coordinates are handled on
           the negative strand.  In the qStart/qEnd fields the
           coordinates are where it matches from the point of view of
           the forward strand (even when the match is on the reverse
           strand). However on the qStarts[] list, the coordinates are
           reversed.

        This class works in forward coordinates for the query and
        forward/reverse coordinates for the sbjct.

        For a negative strand match, the following is done:
           * invert mSbjctFrom and mSbjctTo with mSbjctLength
           * add block sizes to mQueryStarts and mSbjctStarts
           * invert mQueryStarts and mSbjctStarts
           * reverse blocksize, mQueryStarts and mSbjctStarts

        """

        block_sizes = self.mBlockSize
        query_starts = self.mQueryBlockStarts
        sbjct_starts = self.mSbjctBlockStarts

        if self.mSbjctStrand == "-":
            self.mSbjctFrom, self.mSbjctTo = self.mSbjctLength - \
                self.mSbjctTo, self.mSbjctLength - self.mSbjctFrom
            query_starts = [
                self.mQueryLength - x - y for x, y in zip(query_starts, block_sizes)]
            sbjct_starts = [
                self.mSbjctLength - x - y for x, y in zip(sbjct_starts, block_sizes)]
            query_starts.reverse()
            sbjct_starts.reverse()

        self.mQueryStarts = [x - self.mQueryFrom for x in query_starts]
        self.mSbjctStarts = [x - self.mSbjctFrom for x in sbjct_starts]

    def _switchQueryStrandFlag(self):
        '''switch the strand flag of the query
        '''

        if self.strand == "-":
            self.strand = "+"
        else:
            self.strand = "-"

    def switchTargetStrand(self):
        """switch the target strand.

        Use in cases in which a feature has been defined on the
        negative target strand with reverse coordinates. The result
        will be the same alignment using forward coordinates on the
        target.

        This method will also update the query strand and coordinates.
        """

        block_sizes = self.mBlockSizes
        query_starts = self.mQueryBlockStarts
        sbjct_starts = self.mSbjctBlockStarts

        # invert alignment
        self.mQueryBlockStarts = [
            self.mQueryLength - x - y
            for x, y in zip(query_starts, block_sizes)]
        self.mQueryBlockStarts.reverse()
        self.mSbjctBlockStarts = [
            self.mSbjctLength - x - y
            for x, y in zip(sbjct_starts, block_sizes)]
        self.mSbjctBlockStarts.reverse()
        self.mBlockSizes.reverse()

        # update target coordinates only
        self.mSbjctFrom, self.mSbjctTo = self.mSbjctLength - \
            self.mSbjctTo, self.mSbjctLength - self.mSbjctFrom
        self._switchQueryStrandFlag()

    def fromTable(self, data):

        try:
            (nmatches, nmismatches, nrepmatches, nns,
             query_ngaps_counts, query_ngaps_bases,
             sbjct_ngaps_counts, sbjct_ngaps_bases,
             strand,
             query_id, query_length, query_from, query_to,
             sbjct_id, sbjct_length, sbjct_from, sbjct_to,
             nblocks, block_sizes,
             query_block_starts, sbjct_block_starts) = data[:21]
        except ValueError:
            raise ParsingError("parsing error: %i fields" %
                               len(data), "\t".join(data))

        nmatches, nmismatches, nrepmatches, nns = list(map(
            int,
            (nmatches, nmismatches, nrepmatches, nns)))

        self.mNMatches = nmatches
        self.mNMismatches = nmismatches
        self.mNRepMatches = nrepmatches
        self.mNns = nns
        self.mQueryNGapsCounts = int(query_ngaps_counts)
        self.mQueryNGapsBases = int(query_ngaps_bases)
        self.mSbjctNGapsCounts = int(sbjct_ngaps_counts)
        self.mSbjctNGapsBases = int(sbjct_ngaps_bases)
        self.strand = strand
        self.mQueryId = query_id
        self.mQueryLength = int(query_length)
        self.mQueryFrom = int(query_from)
        self.mQueryTo = int(query_to)
        self.mSbjctId = sbjct_id
        self.mSbjctLength = int(sbjct_length)
        self.mSbjctFrom = int(sbjct_from)
        self.mSbjctTo = int(sbjct_to)
        self.mNBlocks = int(nblocks)
        self.mBlockSizes = list(map(int, block_sizes[:-1].split(",")))
        self.mQueryBlockStarts = list(
            map(int, query_block_starts[:-1].split(",")))
        self.mSbjctBlockStarts = list(
            map(int, sbjct_block_starts[:-1].split(",")))

        # this makes sure that the block positions are rescaled
        if self.mQueryLength != 0:
            self.mQueryCoverage = 100.0 * \
                (self.mNMismatches + self.mNMatches) / self.mQueryLength
        else:
            self.mQueryCoverage = 0

        if self.mSbjctLength != 0:
            self.mSbjctCoverage = 100.0 * \
                (self.mNMismatches + self.mNMatches) / self.mSbjctLength
        else:
            self.mSbjctCoverage = 0

        if nmatches + nmismatches > 0:
            self.mPid = 100.0 * float(nmatches) / (nmatches + nmismatches)
        else:
            self.mPid = 100.0

    def fromMaq(self, maq):
        """build BLAT entry from a MAQ match.

        see :class:`Maq.Match`.
        """
        start = maq.start - 1
        l = maq.mLength
        self.mQueryId = maq.mName
        self.mSbjctId = maq.contig
        self.mQueryFrom = 0
        self.mQueryTo = l
        self.mQueryLength = l
        self.strand = maq.strand
        self.mSbjctFrom = start
        self.mSbjctTo = start + l
        self.mNMismatches = maq.mNMismatches
        self.mNMatches = l - self.mNMismatches
        self.mBlockSizes = [l]
        self.mQueryBlockStarts = [0]
        self.mSbjctBlockStarts = [start]
        self.mNBlocks = 1

    def getBlocks(self):
        """return a list of aligned blocks."""
        return list(zip(self.mQueryBlockStarts,
                        self.mSbjctBlockStarts,
                        self.mBlockSizes))

    def __str__(self):
        return "\t".join(map(str, (
            self.mNMatches,
            self.mNMismatches,
            self.mNRepMatches,
            self.mNns,
            self.mQueryNGapsCounts,
            self.mQueryNGapsBases,
            self.mSbjctNGapsCounts,
            self.mSbjctNGapsBases,
            self.strand,
            self.mQueryId,
            self.mQueryLength,
            self.mQueryFrom,
            self.mQueryTo,
            self.mSbjctId,
            self.mSbjctLength,
            self.mSbjctFrom,
            self.mSbjctTo,
            self.mNBlocks,
            ",".join(map(str, self.mBlockSizes)) + ",",
            ",".join(map(str, self.mQueryBlockStarts)) + ",",
            ",".join(map(str, self.mSbjctBlockStarts)) + ",")))

    def getMapQuery2Target(self):
        """return a map between query to target.

        If the strand is "-", the coordinates for query are on
        the negative strand.
        """

        map_query2target = alignlib_lite.py_makeAlignmentBlocks()

        f = alignlib_lite.py_AlignmentFormatBlat(
            "%i\t%i\t%i\t%i\t%s\t%s\t%s\n" % (
                min(self.mQueryBlockStarts),
                max(self.mQueryBlockStarts),
                min(self.mSbjctBlockStarts),
                max(self.mSbjctBlockStarts),
                ",".join([str(x) for x in self.mQueryBlockStarts]) + ",",
                ",".join([str(x) for x in self.mSbjctBlockStarts]) + ",",
                ",".join([str(x) for x in self.mBlockSizes]) + ","))
        f.copy(map_query2target)

        return map_query2target

    def getMapTarget2Query(self):
        """return a map between target to query.

        If the strand is "-", the coordinates for query are on
        the negative strand.
        """

        map_target2query = alignlib_lite.py_makeAlignmentBlocks()

        f = alignlib_lite.py_AlignmentFormatBlat(
            "%i\t%i\t%i\t%i\t%s\t%s\t%s\n" % (
                min(self.mSbjctBlockStarts),
                max(self.mSbjctBlockStarts),
                min(self.mQueryBlockStarts),
                max(self.mQueryBlockStarts),
                ",".join([str(x) for x in self.mSbjctBlockStarts]) + ",",
                ",".join([str(x) for x in self.mQueryBlockStarts]) + ",",
                ",".join([str(x) for x in self.mBlockSizes]) + ","))
        f.copy(map_target2query)
        return map_target2query

    def copy(self):
        return copy.copy(self)

    def fromMap(self, map_query2target, use_strand=None):
        """return a map between query to target."""

        self.mNMatches = map_query2target.getNumAligned()
        f = str(alignlib_lite.py_AlignmentFormatBlat(map_query2target))

        self.mQueryFrom, self.mQueryTo, self.mSbjctFrom, self.mSbjctTo, \
            self.mQueryBlockStarts, self.mSbjctBlockStarts, self.mBlockSizes = f.split(
                "\t")

        if self.mBlockSizes:
            self.mBlockSizes = list(map(int, self.mBlockSizes[:-1].split(",")))
            self.mQueryBlockStarts = list(map(
                int, self.mQueryBlockStarts[:-1].split(",")))
            self.mSbjctBlockStarts = list(map(
                int, self.mSbjctBlockStarts[:-1].split(",")))
        else:
            self.mBlockSizes = []
            self.mQueryBlockStarts = []
            self.mSbjctBlockStarts = []

        self.mNBlocks = len(self.mBlockSizes)

        self.mQueryFrom, self.mQueryTo, self.mSbjctFrom, self.mSbjctTo = \
            list(map(int,
                     (self.mQueryFrom, self.mQueryTo,
                      self.mSbjctFrom, self.mSbjctTo)))

        # queryfrom and queryto are always forward strand coordinates
        if use_strand and self.strand == "-":
            self.mQueryFrom, self.mQueryTo = self.mQueryLength - \
                self.mQueryTo, self.mQueryLength - self.mQueryFrom

    def fromPair(self,
                 query_start, query_size, query_strand, query_seq,
                 target_start, target_size, target_strand, target_seq):
        '''fill from two aligned sequences.

        Note that sequences are case-sensitive.'''

        self.mQueryLength = query_size
        self.mSbjctLength = target_size

        map_query2target = alignlib_lite.py_makeAlignmentBlocks()

        assert len(query_seq) == len(target_seq)

        x, y = query_start, target_start
        nmatches, nmismatches = 0, 0
        for q, t in zip(query_seq, target_seq):
            tq, tt = q != "-", t != "-"
            if tq and tt:
                map_query2target.addPair(x, y)
                if q == t:
                    nmatches += 1
                else:
                    nmismatches += 1

            if tq:
                x += 1
            if tt:
                y += 1

        self.mNMatches, self.mNMismatches = nmatches, nmismatches
        self.strand = query_strand
        # the following call will set query_from, query_to for the forward strand
        # though block coordinates might be on the negative strand
        self.fromMap(map_query2target, use_strand=True)

        # if target is on negative strand, swop strands
        if target_strand == "-":
            # swap target strand - this will also swap the query strand
            self.switchTargetStrand()

    def iterator_introns(self):

        b = self.getBlocks()
        last_tend = b[0][1] + b[0][2]
        for qstart, tstart, size in b[1:]:
            yield (last_tend, tstart)
            last_tend = tstart + size
        raise StopIteration

    def iterator_exons(self):

        for qstart, tstart, size in self.getBlocks():
            yield tstart, tstart + size
        raise StopIteration

    def iterator_introns(self):

        b = self.getBlocks()
        last_tend = b[0][1] + b[0][2]
        for qstart, tstart, size in b[1:]:
            yield (last_tend, tstart)
            last_tend = tstart + size
        raise StopIteration

    def iterator_query_exons(self):

        for qstart, tstart, size in self.getBlocks():
            yield qstart, qstart + size
        raise StopIteration

    def iterator_sbjct_exons(self):

        for qstart, tstart, size in self.getBlocks():
            yield tstart, tstart + size
        raise StopIteration

    def iterator_query_introns(self):

        b = self.getBlocks()
        last_qend = b[0][0] + b[0][2]
        for qstart, tstart, size in b[1:]:
            yield (last_qend, qstart)
            last_qend = qstart + size
        raise StopIteration

    def getHeader(self):

        return HEADER

    def getHeaders(self):
        return ("match", "mismatch", "repeats", "Ns",
                "qGapCount", "qGapBases", "tGapCount", "tGapBases", "strand",
                "qName", "qSize", "qStart", "qEnd",
                "tName", "tSize", "tStart", "tEnd",
                "blockCount", "blockSizes",
                "qStarts", "tStarts")


class MatchPSLX(Match):

    def __init__(self):
        Match.__init__(self)
        self.mQuerySequence = []
        self.mSbjctSequence = []

    def __str__(self):
        return "\t".join((Match.__str__(self),
                          ",".join(self.mQuerySequence) + ",",
                          ",".join(self.mSbjctSequence) + ","))

    def fromTable(self, data):

        Match.fromTable(self, data)

        try:
            query_sequence, sbjct_sequence = data[21:23]

        except ValueError:
            raise ParsingError("parsing error", "\t".join(data))

        self.mQuerySequence = query_sequence[:-1].split(",")
        self.mSbjctSequence = sbjct_sequence[:-1].split(",")

    def getHeaders(self):
        return Match.getHeaders(self) + ("qSequence", "tSequence")

    def fromPSL(self, other, query_sequence, sbjct_sequence):
        """fill entry from a psl match.

        sequences are on forward strand starting at
        query_from and sbjct_from, respectively.
        """

        (self.mNMatches,
         self.mNMismatches,
         self.mNRepMatches,
         self.mNns,
         self.mQueryNGapsCounts,
         self.mQueryNGapsBases,
         self.mSbjctNGapsCounts,
         self.mSbjctNGapsBases,
         self.strand,
         self.mQueryId,
         self.mQueryLength,
         self.mQueryFrom,
         self.mQueryTo,
         self.mSbjctId,
         self.mSbjctLength,
         self.mSbjctFrom,
         self.mSbjctTo,
         self.mNBlocks,
         self.mBlockSizes,
         self.mQueryBlockStarts,
         self.mSbjctBlockStarts) = (other.mNMatches,
                                    other.mNMismatches,
                                    other.mNRepMatches,
                                    other.mNns,
                                    other.mQueryNGapsCounts,
                                    other.mQueryNGapsBases,
                                    other.mSbjctNGapsCounts,
                                    other.mSbjctNGapsBases,
                                    other.strand,
                                    other.mQueryId,
                                    other.mQueryLength,
                                    other.mQueryFrom,
                                    other.mQueryTo,
                                    other.mSbjctId,
                                    other.mSbjctLength,
                                    other.mSbjctFrom,
                                    other.mSbjctTo,
                                    other.mNBlocks,
                                    other.mBlockSizes,
                                    other.mQueryBlockStarts,
                                    other.mSbjctBlockStarts)

        if self.strand == "-":
            query_sequence = string.translate(
                query_sequence, string.maketrans("ACGTacgt", "TGCAtgca"))[::-1]

        self.mQuerySequence = []
        self.mSbjctSequence = []
        offset_query = min(self.mQueryBlockStarts)
        offset_sbjct = min(self.mSbjctBlockStarts)

        for start_query, start_sbjct, size in other.getBlocks():
            self.mQuerySequence.append(
                query_sequence[start_query - offset_query:start_query - offset_query + size])
            self.mSbjctSequence.append(
                sbjct_sequence[start_sbjct - offset_sbjct:start_sbjct - offset_sbjct + size])


def _iterate(infile):
    """iterator over psl output.

    The header is optional.
    """

    format = "ps3"

    while 1:
        line = infile.readline()
        if not line:
            raise StopIteration

        if line[0] == "#":
            continue

        if line.startswith("match"):
            continue

        if line.startswith("psLayout version 3"):
            format = "ps3"
            for x in range(4):
                infile.readline()
            continue

        match = Match()
        match.fromTable(line[:-1].split("\t"))
        yield match


class BlatIterator:

    def __init__(self, f, *args, **kwargs):
        self.mIterator = _iterate(f)

    def __iter__(self):
        return self

    def __next__(self):
        try:
            return next(self.mIterator)
        except StopIteration:
            return None

    def next(self):
        return self.__next__()


def iterator2(infile):
    """iterate over the contents of a psl file.
    """

    while 1:
        line = infile.readline()
        if not line:
            raise StopIteration
        if line[0] == "#":
            continue
        if line.startswith("match"):
            continue
        if line.startswith("psLayout version 3"):
            for x in range(4):
                infile.readline()
            continue
        match = Match()
        # match.fromTable( line[:-1].split() )
        yield match


def iterator(infile):
    """iterate over the contents of a psl file.
    """
    while 1:
        line = infile.readline()
        if not line:
            raise StopIteration
        if line[0] == "#":
            continue
        if line.startswith("match"):
            continue
        if line.startswith("psLayout version 3"):
            for x in range(4):
                infile.readline()
            continue
        match = Match()
        match.fromTable(line[:-1].split())
        yield match


def iterator_pslx(infile):
    """iterate over the contents of a pslx file.
    """
    while 1:
        line = infile.readline()
        if not line:
            raise StopIteration
        if line[0] == "#":
            continue
        if line.startswith("match"):
            continue
        if line.startswith("psLayout version 3"):
            for x in range(4):
                infile.readline()
            continue
        match = MatchPSLX()
        match.fromTable(line[:-1].split())
        yield match


def iterator_target_overlap(infile, merge_distance):
    '''iterate over psl formatted infile and return
    blocks of target overlapping alignments.'''

    last_sbjct_id = None
    start = None
    end = None
    matches = []
    processed_contigs = set()

    while 1:

        match = next(infile)

        if match is None:
            break

        if match.mSbjctId != last_sbjct_id or match.mSbjctFrom >= (end + merge_distance):
            if last_sbjct_id:
                yield matches
            matches = []
            if last_sbjct_id != match.mSbjctId and match.mSbjctId in processed_contigs:
                raise ValueError("input not sorted by target (contig,start): already encountered %s\n%s" %
                                 (match.mSbjctId, str(match)))

            processed_contigs.add(last_sbjct_id)

            last_sbjct_id = match.mSbjctId
            start = match.mSbjctFrom
            end = match.mSbjctTo

        if match.mSbjctFrom < start:
            raise ValueError("input not sorted by target (contig,start): %i < %i\n%s" %
                             (match.mSbjctFrom, start, str(match)))

        end = max(match.mSbjctTo, end)
        matches.append(match)

    if last_sbjct_id:
        yield matches


def iterator_query_overlap(infile, merge_distance):
    '''iterate over psl formatted infile and return
    blocks of target overlapping alignments.'''

    last_query_id = None
    start, end = None, None
    matches = []
    processed_contigs = set()

    while 1:

        match = next(infile)

        if match is None:
            break

        if match.mQueryId != last_query_id or match.mQueryFrom >= (end + merge_distance):
            if last_query_id:
                yield matches
            matches = []

            if last_query_id != match.mQueryId and match.mQueryId in processed_contigs:
                raise ValueError("input not sorted by query (contig,start): already encountered %s\n%s" %
                                 (match.mQueryId, str(match)))

            processed_contigs.add(last_query_id)

            last_query_id = match.mQueryId
            start, end = match.mQueryFrom, match.mQueryTo

        if match.mQueryFrom < start:
            raise ValueError("input not sorted by query (contig,start): %i < %i\n%s" %
                             (match.mQueryFrom, start, str(match)))

        end = max(match.mQueryTo, end)
        matches.append(match)

    if last_query_id:
        yield matches


def iterator_test(infile, report_step=100000):
    '''only output parseable lines from infile.'''

    ninput, noutput, nerrors = 0, 0, 0

    while 1:
        try:
            x = next(infile)
        except ParsingError as msg:
            nerrors += 1
            ninput += 1
            E.warn(str(msg))
            continue
        except StopIteration:
            break

        if not x:
            break

        ninput += 1

        if ninput % report_step == 0:
            E.info("progress: ninput=%i, noutput=%i" % (ninput, noutput))

        yield x
        noutput += 1

    E.info("iterator_test: ninput=%i, noutput=%i, nerrors=%i" %
           (ninput, noutput, nerrors))


def iterator_per_query(iterator_psl):
    """iterate over the contents of a psl file per query
    """

    data = collections.defaultdict(list)

    for match in iterator_psl:
        data[match.mQueryId].append(match)

    for x in list(data.values()):
        yield x

    raise StopIteration

FIELDS = ("matches",
          "misMatches",
          "repMatches",
          "nCount",
          "qNumInsert",
          "qBaseInsert",
          "tNumInsert",
          "tBaseInsert",
          "strand",
          "qName",
          "qSize",
          "qStart",
          "qEnd",
          "tName",
          "tSize",
          "tStart",
          "tEnd",
          "blockCount",
          "blockSizes",
          "qStarts",
          "tStarts")


def addAlignments(matches, shift=0, by_query=False):
    """building a genome to query alignment for all matches

    The genome alignment is shifted by *shift*.
    """

    if by_query:
        for match in matches:
            if not hasattr(match, "mMapQuery2Target"):
                map_query2target = match.getMapQuery2Target()
                if shift:
                    map_query2target.moveAlignment(shift, 0)
                match.mMapQuery2Target = map_query2target
    else:
        for match in matches:
            if not hasattr(match, "mMapTarget2Query"):
                map_target2query = match.getMapTarget2Query()
                if shift:
                    map_target2query.moveAlignment(shift, 0)
                match.mMapTarget2Query = map_target2query


def getComponents(matches, max_distance=0, min_overlap=0, by_query=False):
    """return overlapping matches.

    max_distance
       allow reads to be joined if they are # residues apart.
       Adjacent reads are 1 residue apart, overlapping reads are 0 residues
       apart
    min_overlap
       require at least # residues to be overlapping

    """

    addAlignments(matches, by_query=by_query)

    components = Components.IComponents()

    for x in range(0, len(matches)):
        components.add(x, x)

    if min_overlap > 0 and max_distance > 0:
        raise ValueError(
            "both min_overlap (%i) and max_distance (%i) > 0" % (min_overlap, max_distance))

    if by_query:
        if min_overlap > 0:
            f = lambda x, y: alignlib_lite.py_getAlignmentOverlap(
                matches[x].mMapQuery2Target,
                matches[y].mMapQuery2Target,
                alignlib_lite.py_RR) >= min_overlap
        else:
            f = lambda x, y: alignlib_lite.py_getAlignmentShortestDistance(
                matches[x].mMapQuery2Target,
                matches[y].mMapQuery2Target,
                alignlib_lite.py_RR) <= max_distance
    else:
        if min_overlap > 0:
            f = lambda x, y: alignlib_lite.py_getAlignmentOverlap(
                matches[x].mMapTarget2Query,
                matches[y].mMapTarget2Query,
                alignlib_lite.py_RR) >= min_overlap
        else:
            f = lambda x, y: alignlib_lite.py_getAlignmentShortestDistance(
                matches[x].mMapTarget2Query,
                matches[y].mMapTarget2Query,
                alignlib_lite.py_RR) <= max_distance

    for x in range(len(matches)):
        for y in range(0, x):
            if f(x, y):
                components.add(x, y)

    return components.getComponents()



"""
fasta2bed.py - segment sequences
================================

:Tags: Genomics Sequences Intervals FASTA BED Conversion

Purpose
-------

This script takes a genomic sequence in :term:`fasta` format
and applies various segmentation algorithms.

The methods implemented (``--methods``) are:

GCProfile
   use `GCProfile <http://tubic.tju.edu.cn/GC-Profile/>`_ to segment by
   G+C content. GCProfile needs to be installed and in the PATH.

cpg
   output all locations of cpg in the genome

fixed-width-windows-gc
   output fixed width windows of a certain size adding their
   G+C content as score

gaps
   ouput all locations of assembly gaps (blocks of `N`)
   in the genomic sequences

ungapped
   output ungapped locations in the genomic sequences

Usage
-----

Type::

   python fasta2bed.py --method=gap < in.fasta > out.bed


Type::

   python fasta2bed.py --help

for command line help.

Command line options
--------------------

"""

import os
import sys
import re
import tempfile
import subprocess
import glob
import collections
import pybedtools
import CGAT.Experiment as E
import CGAT.IOTools as IOTools
import CGAT.FastaIterator as FastaIterator
import CGAT.IOTools as IOTools

EXECUTABLE = "GCProfile"


def segmentWithCpG(infile, with_contig_sizes=False):
    '''segment a fasta file, output locations of CpG.'''

    ninput, nskipped, noutput = 0, 0, 0

    iterator = FastaIterator.FastaIterator(infile)

    segments, contig_sizes = [], collections.OrderedDict()

    for cur_record in iterator:
        ninput += 1
        contig = re.sub("\s.*", "", cur_record.title)
        last = None
        contig_sizes[contig] = (0, len(cur_record.sequence))
        for pos, this in enumerate(cur_record.sequence.upper()):
            if last == "C" and this == "G":
                segments.append((contig, pos - 1, pos + 1, 1.0))
            last = this

    E.info("ninput=%i, noutput=%i, nskipped=%i" % (ninput, noutput, nskipped))

    if with_contig_sizes:
        return segments, contig_sizes

    return segments


def segmentWindowsCpG(infile, window_size=100, min_cpg=1):
    '''segment a fasta file based on the locations of CpG.

    Locate all CpG in sequences and centre windows of size *window_size*
    around them. Merge all windows and keep all with *min_cpg* CpG.
    '''

    cpgs, contig_sizes = segmentWithCpG(infile, with_contig_sizes=True)

    # save cpgs to temporary file
    tempf = tempfile.NamedTemporaryFile(mode="w", delete=False)
    tempf.write("\n".join(["%s\t%i\t%i\n" % (contig, start, end)
                           for contig, start, end, gc in cpgs]) + "\n")
    tempf.close()

    cpgs = pybedtools.BedTool(tempf.name)
    cpgs.set_chromsizes(contig_sizes)
    extended = cpgs.slop(b=window_size // 2)
    merged = extended.merge(o="count", c=3)
    filtered = merged.filter(lambda x: int(x.name) >= min_cpg)

    os.unlink(tempf.name)

    # return CpG content (not C+C content)
    return [(x.chrom, x.start, x.stop, float(x.name) / (x.stop - x.start) / 2)
            for x in filtered]


def segmentWithGCProfile(infile, options):
    '''segment a fasta file with GCProfile.'''

    ninput, nskipped, noutput = 0, 0, 0

    iterator = FastaIterator.FastaIterator(infile)

    tmpdir = tempfile.mkdtemp()
    E.info("working in %s" % tmpdir)

    while 1:
        try:
            cur_record = next(iterator)
        except StopIteration:
            break

        if cur_record is None:
            break
        ninput += 1

        contig = re.sub("\s.*", "", cur_record.title)
        filename = os.path.join(tmpdir, contig) + ".fasta"
        with IOTools.openFile(filename, "w") as f:
            f.write(">%s\n%s\n" % (contig, cur_record.sequence))

        E.info("running %s on %s" % (EXECUTABLE, contig))

        statement = "%s %s -t %i -g %f -i %i " % \
                    (EXECUTABLE, filename,
                     options.gcprofile_halting_parameter,
                     options.gcprofile_gap_size,
                     options.gcprofile_min_length)

        E.debug("statement = %s" % statement)

        process = subprocess.Popen(statement,
                                   cwd=tmpdir,
                                   shell=True,
                                   stdin=None,
                                   stdout=subprocess.PIPE,
                                   stderr=subprocess.PIPE)

        stdout, stderr = process.communicate()

        E.debug("GProfile: stdout=%s" % stdout)
        E.debug("GProfile: stderr=%s" % stderr)

        # process.stdin.close()
        if process.returncode != 0:
            raise OSError("Child was terminated by signal %i: \nThe stderr was: \n%s\n%s\n" %
                          (-process.returncode, stderr, statement))

        noutput += 1

    segments = []

    for f in glob.glob(os.path.join(tmpdir, "*.SegGC")):
        contig = os.path.basename(f[:-len(".SegGC")])
        with IOTools.openFile(f, "r") as infile:
            start = None
            for line in infile:
                pos, gc = line[:-1].split("\t")
                try:
                    gc = float(gc)
                except ValueError:
                    gc = None
                end = int(pos)
                # ignore the double lines
                if gc is not None and start is not None and end - start > 2:
                    segments.append((contig, start, end, gc))
                start = int(pos) - 1

    E.info("ninput=%i, noutput=%i, nskipped=%i" % (ninput, noutput, nskipped))

    # shutil.rmtree( tmpdir )

    return segments


def segmentFixedWidthWindows(infile, window_size, window_shift):
    """return a list of fixed contig sizes."""

    ninput, nskipped, noutput = 0, 0, 0

    iterator = FastaIterator.FastaIterator(infile)
    window_shift = window_size
    # at most 50% can be gap
    gap_cutoff = int(window_size // 2)
    segments = []

    while 1:
        ninput += 1
        try:
            cur_record = next(iterator)
        except StopIteration:
            break

        if cur_record is None:
            break
        contig = re.sub("\s.*", "", cur_record.title)
        seq = cur_record.sequence
        size = len(cur_record.sequence)

        for x in range(0, size, window_shift):
            s = seq[x:x + window_size].upper()
            gc, at = 0, 0
            for c in s:
                if c in "GC":
                    gc += 1
                elif c in "AT":
                    at += 1

            # skip segments containing mostly gaps
            if window_size - (gc + at) > gap_cutoff:
                nskipped += 1
                continue

            segments.append(
                (contig, x, x + window_size, float(gc) / (gc + at)))
        noutput += 1

    E.info("ninput=%i, noutput=%i, nskipped_windows=%i" %
           (ninput, noutput, nskipped))

    return segments


def gapped_regions(seq, gap_chars):
    '''iterator yielding gapped regions in seq.'''
    is_gap = seq[0] in gap_chars
    last = 0
    size = len(seq)
    for x, c in enumerate(seq):
        if c in gap_chars:
            if not is_gap:
                last = x
                is_gap = True
        else:
            if is_gap:
                yield(last, x)
                last = x
                is_gap = False
    if is_gap:
        yield last, size


def segmentGaps(infile, gap_char):

    iterator = FastaIterator.FastaIterator(infile)

    while 1:
        try:
            cur_record = next(iterator)
        except StopIteration:
            break

        if cur_record is None:
            break
        contig = re.sub("\s.*", "", cur_record.title)

        for start, end in gapped_regions(cur_record.sequence, gap_char):
            yield(contig, start, end, 0)


def segmentUngapped(infile, gap_char, min_gap_size=0):

    iterator = FastaIterator.FastaIterator(infile)

    while 1:
        try:
            cur_record = next(iterator)
        except StopIteration:
            break

        if cur_record is None:
            break
        contig = re.sub("\s.*", "", cur_record.title)
        size = len(cur_record.sequence)

        last_end = 0
        for start, end in gapped_regions(cur_record.sequence, gap_char):
            if end - start < min_gap_size:
                continue

            if last_end != 0:
                yield(contig, last_end, start, 0)
            last_end = end

        if last_end < size:
            yield(contig, last_end, size, 0)


def main(argv=None):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv:
        argv = sys.argv

    # setup command line parser
    parser = E.OptionParser(
        version="%prog version: $Id",
        usage=globals()["__doc__"])

    parser.add_option(
        "--gcprofile-gap-size", dest="gcprofile_gap_size",
        type="float",
        help="GCProfile: Choose n as the gap size to be filtered "
        "(If n > 1, n bp is set as the gap size to be filtered. "
        "If 0 < n < 1, for example, n = 0.01, means gaps less than 1% "
        "of the input sequence length will be filtered "
        "[default=%default].")

    parser.add_option(
        "--gcprofile-min-length", dest="gcprofile_min_length",
        type="int",
        help="GCProfile: minimum length [default=%default]")

    parser.add_option(
        "--gcprofile-halting-parameter",
        dest="gcprofile_halting_parameter", type="int",
        help="GCProfile: halting parameter [default=%default]")

    parser.add_option(
        "-m", "--method", dest="method", type="choice",
        choices=(
            "GCProfile",
            "fixed-width-windows-gc",
            "cpg",
            "windows-cpg",
            "gaps",
            "ungapped",
            "windows"),
        help="Method to use for segmentation [default=%default]")

    parser.add_option(
        "-w", "--window-size=", dest="window_size",
        type="int",
        help="window size for fixed-width windows [default=%default].")

    parser.add_option(
        "-s", "--window-shift=", dest="window_shift", type="int",
        help="shift size fixed-width windows [default=%default].")

    parser.add_option(
        "--min-cpg", dest="min_cpg", type="int",
        help="minimum number of CpG for windows-cpg [default=%default]")

    parser.add_option(
        "--min-interval-length", dest="min_length", type="int",
        help="minimum length for ungapped regions [default=%default]")

    parser.set_defaults(
        gcprofile_halting_parameter=1000,
        gcprofile_gap_size=0.01,
        gcprofile_min_length=3000,
        window_size=10000,
        method="GCProfile",
        gap_char="NnXx",
        min_length=0,
        window_shift=10000,
        min_cpg=1,
    )

    # add common options (-h/--help, ...) and parse command line
    (options, args) = E.Start(parser, argv=argv)

    # do sth

    if options.method == "GCProfile":
        segments = segmentWithGCProfile(options.stdin, options)
    elif options.method == "cpg":
        segments = segmentWithCpG(options.stdin)
    elif options.method == "windows-cpg":
        segments = segmentWindowsCpG(options.stdin,
                                     options.window_size,
                                     options.min_cpg)
    elif options.method == "Isoplotter":
        segments = segmentWithIsoplotter(options.stdin, options)
    elif options.method == "fixed-width-windows-gc":
        segments = segmentFixedWidthWindows(options.stdin,
                                            options.window_size,
                                            options.window_shift,
                                            )
    elif options.method == "gaps":
        segments = segmentGaps(options.stdin, options.gap_char)
    elif options.method == "ungapped":
        segments = segmentUngapped(
            options.stdin, options.gap_char, options.min_length)
    else:
        raise ValueError("unknown method %s" % (method))
    x = 0
    for contig, start, end, gc in segments:
        x += 1
        options.stdout.write("%s\n" % "\t".join(
            (contig, str(start), str(end), str(x), "%6.4f" % (100.0 * gc))))

    # write footer and output benchmark information.
    E.Stop()

if __name__ == "__main__":
    sys.exit(main(sys.argv))



'''split_gff - split a gff file into chunks
=============================================

:Tags: Genomics Intervals Genesets GFF Manipulation

Purpose
-------

Split gff file into chunks. Overlapping entries will always be output
in the same chunk. Input is read from stdin unless otherwise
specified. The input needs to be contig/start position sorted.

Options
-------

-i --min-chunk-size

    This option specifies how big each chunck should
    be, in terms of the number of gff lines to be
    included. Because overlapping lines are always
    output to the same file, this should be considered
    a minimum size.

-n, --dry-run

    This options tells the script not to actaully write
    any files, but it will output a list of the files
    that would be output.

Example
-------

    cgat splitgff -i 1 < in.gff

where in.gff looks like:

    chr1	.	exon	1	10	.	+	.
    chr1	.	exon	8	100	.	+	.
    chr1	.	exon	102	150	.	+	.

will produce two files that look like:

    000001.chunk:
    chr1	.	exon	1	10	.	+	.
    chr1	.	exon	8	100	.	+	.

    000002.chunk:
    chr1	.	exon	102	150	.	+	.

Usage
-----

   cgat splitgff [OPTIONS]

Will read a gff file from stdin and split into multiple gff files.

   cgat split_gff -I GFF [OPTIONS]

Will read the gff file GFF and split into multiple gff files.

Command line options
--------------------

'''

import sys
import os
import CGAT.GTF as GTF
import CGAT.IOTools as IOTools
import CGAT.Experiment as E


class OutputChunk:

    def __init__(self, output_filename_pattern, dry_run=False):
        self.nchunk = 0
        self.output_filename_pattern = output_filename_pattern
        self.dry_run = dry_run

    def createOpen(self, mode="w", header=None):
        """open file. Check first, if directory exists.
        """

        self.nchunk += 1
        filename = self.output_filename_pattern % self.nchunk

        if self.dry_run:
            E.info("opening file %s" % filename)
            returnIOTools.openFile("/dev/null", mode)

        if mode in ("w", "a"):
            dirname = os.path.dirname(filename)
            if dirname and not os.path.exists(dirname):
                os.makedirs(dirname)

        if os.path.exists(filename):
            existed = True
        else:
            existed = False

        f = IOTools.openFile(filename, mode)

        if header and not existed:
            f.write(header + "\n")

        return f

    def __call__(self, chunk):
        """output a chunk into a new file."""
        outfile = self.createOpen()
        for c in chunk:
            outfile.write(str(c) + "\n")
        outfile.close()
        return len(chunk)


def main(argv=None):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if argv is None:
        argv = sys.argv

    parser = E.OptionParser(
        version="%prog version: $Id$",
        usage=globals()["__doc__"])

    parser.add_option(
        "-i", "--min-chunk-size", dest="min_chunk_size", type="int",
        help="minimum chunk size [default=%default].")

    parser.add_option(
        "-n", "--dry-run", dest="dry_run", action="store_true",
        help="do not create any files [default=%default].")

    parser.set_defaults(
        method="overlap",
        dry_run=False,
        min_chunk_size=2,
        output_filename_pattern="%06i.chunk",
    )

    (options, args) = E.Start(parser, add_output_options=True)

    gffs = GTF.iterator(options.stdin)

    ninput, noutput, nchunks = 0, 0, 0

    outputChunk = OutputChunk(options.output_filename_pattern,
                              dry_run=options.dry_run)

    if options.method == "overlap":

        last_contig, last_to = None, 0
        chunk = []
        for gff in gffs:
            ninput += 1
            if len(chunk) >= options.min_chunk_size and \
                    (gff.contig != last_contig or
                     gff.start > last_to):
                noutput += outputChunk(chunk)
                nchunks += 1
                chunk = []
                last_contig, last_to = gff.contig, gff.end

            chunk.append(gff)
            last_to = max(gff.end, last_to)

        noutput += outputChunk(chunk)
        nchunks += 1

    E.info("ninput=%i, noutput=%i, nchunks=%i" % (ninput, noutput, nchunks))

    E.Stop()

if __name__ == "__main__":
    sys.exit(main(sys.argv))




import sys
import os
import shutil
import CGAT.Experiment as E
import CGAT.IOTools as IOTools
import CGAT.GTF as GTF

'''
Filters a GTF to generate a geneset over which to count reads and look for
differential expression.

Example:
This would keep only lines from the gtf where gene_biotype is pseudogene or
protein_coding, contig is not chrM or chrUn and transcript_support is greater
than 3.

python makeGeneset.py -i refcoding.gtf.gz - o outfile.gtf.gz
-f "gene_biotype=pseudogene+protein_coding" -f "transcript_support-morethan-3"
-c "chrM|chrUn"

GTF "columns" are named, in order
0 contig
1 source
2 feature
3 start
4 end
5 ?
6 strand
7 frame
plus anything before an "=" sign in column 8, e.g. gene_source, exon_number

Options
-g --gtf
path to local gtf to filter

-p --gtfpath
path to online gtf to download and filter

-o --outfile
path to outfile

-c --remove-contigs
Followed by a list of contigs to remove delimited by |
e.g. -c chrMT|chrALT removes all rows of the GTF from chrMT and chrALT

-k --keep-contigs
Followed by a list of contigs to keep delimited by |
e.g. -k chr1|chr12|chr3 only keeps rows of the GTF from chr1, chr12 and chr3

-f --filter
Any number of filters can be specified as -f or --filter followed by a string
laid out as follows:

To keep only lines with values from a list in a specified column:
-f COLUMN=VALUE1+VALUE2+VALUE3
e.g.
-f 'source=protein_coding+lincRNA' will return only rows where source =
   protein_coding or source = lincRNA

To keep only lines with values not in a list in a specified column
-f COLUMN!=VALUE1+VALUE2+VALUE3
e.g.
-f 'source!=protein_coding+lincRNA' will return only rows where source !=
   protein_coding and source != lincRNA

To keep only lines with values listed in a file
-f COLUMN-in_file-FILENAME
e.g.
-f 'source-in_file-a.tsv' will read the file a.tsv and return only rows of
the gtf where source is one of the values in this file

To keep only lines with values not listed in a file
-f COLUMN-notin_file-FILENAME
e.g.
-f 'source-notin_file-a.tsv' will read the file a.tsv and return only rows of
the gtf where source is not one of the values in this file

To keep only lines with a value greater than or less than a threshold in
a specified column
-f COLUMN-morethan-VALUE or COLUMN-lessthan-VALUE
e.g. -f "transcript_support-morethan-3" will return only rows with transcript
support greater than 3

'''

# the old pipeline could also do the following
# but these were only used by cufflinks - do we still want them?
# merge exons seperated by small introns
# ignore very long introns
# remove repetitive RNA - we don't understand what this is?


def getGTF(path):
    statement = "wget %(path)s" % locals()
    os.system(statement)


def filterGTF(gtf, filterstring, tempout):

    if "!=" in filterstring:
        column, value = filterstring.split("!=")
        value = value.split("+")
        filtertype = "notin"

    elif "=" in filterstring:
        column, value = filterstring.split("=")
        value = value.split("+")
        filtertype = "in"

    elif "-in_file-" in filterstring:
        column, value = filterstring.split("-in_file-")
        value = [line.strip() for line in IOTools.openFile(value)]
        filtertype = "in_file"

    elif "-notin_file-" in filterstring:
        column, value = filterstring.split("-notin_file-")
        value = [line.strip() for line in IOTools.openFile(value)]
        filtertype = "notin_file"

    elif "-morethan-" in filterstring:
        column, value = filterstring.split("-morethan-")
        value = float(value)
        filtertype = "morethan"

    elif "-lessthan-" in filterstring:
        column, value = filterstring.split("-lessthan-")
        value = float(value)
        filtertype = "lessthan"

    gfile = IOTools.openFile(gtf)
    G = GTF.iterator(gfile)

    out = IOTools.openFile(tempout, "w")
    for item in G:
        D = item.asDict()
        D['contig'] = item.contig
        D['source'] = item.source
        D['feature'] = item.feature
        D['start'] = item.start
        D['end'] = item.end
        D['strand'] = item.strand
        D['frame'] = item.frame

        if filtertype == "in" or filtertype == 'in_file':
            if D[column] in value:
                out.write("%s\n" % str(item))
        elif filtertype == "notin" or filtertype == 'notin_file':
            if D[column] not in value:
                out.write("%s\n" % str(item))
        elif filtertype == "morethan":
            if float(D[column]) > value:
                out.write("%s\n" % str(item))
        elif filtertype == "lessthan":
            if float(D[column]) < value:
                out.write("%s\n" % str(item))
    out.close()
    gfile.close()


def removeNamedContigs(contigs):
    return """awk '$1 !~ /(%(contigs)s)$/' |""" % locals()


def keepOnlyNamedContigs(contigs):
    return """awk '$1 ~ /(%(contigs)s)$/' |""" % locals()


def main(argv=None):
    if argv is None:
        argv = sys.argv

    parser = E.OptionParser(
        version="%prog version: $Id$",
        usage=globals()["__doc__"])

    parser.add_option("-g", "--gtf", dest="gtf",
                      type="string",
                      help="path to input gtf")

    parser.add_option("-p", "--gtfpath", dest="gtfpath",
                      type="string",
                      help="path to online gtf")

    parser.add_option("-c", "--remove-contigs", dest="remove_contigs",
                      type="string",
                      help="contigs to remove, delimited by |")

    parser.add_option("-k", "--keep-contigs", dest="keep_contigs",
                      type="string",
                      help="""all contigs to keep, delimited by |.  Contigs
                      specified in --remove-contigs will still be removed""")

    parser.add_option("-o", "--outfile", dest="outfile",
                      type="string",
                      help="path to processed output gtf")

    parser.add_option("-f", "--filter", dest="filters",
                      type="string",
                      action="append",
                      help="""List of filters to apply to your GTF""")

    parser.set_defaults(
        remove_contigs=None,
        keep_contigs=None,
    )

    (options, args) = E.Start(parser)

    if options.gtf:
        gtf = options.gtf
    elif options.gtfpath:
        getGTF(options.gtfpath)
        gtf = options.gtfpath.split("/")[-1]
    else:
        raise ValueError("Please provide a GTF or the path to an online GTF")

    if not options.outfile:
        raise ValueError("Please provide an output file name")

    d = 0
    if options.remove_contigs or options.keep_contigs:
        d += 1
        statement = 'zcat %s |' % gtf

        if options.remove_contigs:
            statement += removeNamedContigs(options.remove_contigs)

        if options.keep_contigs:
            statement += keepOnlyNamedContigs(options.keep_contigs)

        if options.outfile.endswith(".gz"):
            outfile = options.outfile
        else:
            outfile = options.outfile + ".gz"

        statement += "gzip > %s " % outfile

        os.system(statement)

    T1 = gtf
    if options.filters:
        d += 1
        for filterstring in options.filters:
            T2 = P.getTempFilename(".")
            T2 = T2 + ".gtf"
            filterGTF(T1, filterstring, T2)
            T1 = T2

        shutil.move(T2, options.outfile)

    if d == 0:
        raise ValueError("No filters provided")


if __name__ == "__main__":
    sys.exit(main(sys.argv))



'''bam2peakshape.py - compute peak shape features from a bam-file
==============================================================

:Tags: Genomics NGS Intervals BAM BED Summary

Purpose
-------

This script takes a :term:`bed` formatted file with regions of
interest, for example binding intervals from a ChIP-Seq
experiment. Using a collection of aligned reads is a :term:`bam`
formatted file or :term:`bigwig` formatted file, the script outputs a
collection of features describing the peak shape.

This script is designed with a slight emphasis on ChIP-Seq datasets.
The main reason that this script is better suited for ChIP-Seq is
that(1) it is able to center the counting window at the summit of
every individual peak; (2) it is also able to use the control ChIP-Seq
library to enable side-by-side comparison of treatment vs control;(3)
it can randomly shift the set of input regions to generate a
artificial set of regions, in the absence of real ChIP-Seq control
library, the random regions can provide a peaks profile that can be
used as the control.

For example, given the peaks regions defined by analyzing some
ChIP-Seq dataset (e.g. by using MACS), and without the need to use any
additional genomic annotations (e.g. ENSEMBL, refseq), we can
visualise the binding profiles of transcriptionfactors ChIP-Seq data
relative to the center of each peak regions.

The script outputs a tab-separated table on stdout containing features
for each interval. A peak is defined as the location of the highest
density in an interval. The width of the peak (peak_width) is defined
as the region around the peak in which the density does not drop below
a threshold of peak_heigt * 90%.

Usage
-----

Detailed usage example
++++++++++++++++++++++

The following command will generate the peak shape plot for the peak
regions defined in :file:`onepeak.bed`, using the reads stored in
:file:`small.bam`.  The command will also create a profile for the
control library.  The control library in this example is re-using the
same reads file :file:`small.bam`, however, in your actual experiment,
it should be a different library (the input library for this ChIP-Seq
experiment).::

    python ./scripts/bam2peakshape.py \
        ./tests/bam2peakshape.py/small.bam \
        ./tests/bam2peakshape.py/onepeak.bed \
        --control-bam-file=./tests/bam2peakshape.py/small.bam \
        --use-interval \
        --normalize-transcript


Output files
++++++++++++

Among the features output are:

+-------------------+---------------------------------------------------------+
|*Column*           |*Content*                                                |
+-------------------+---------------------------------------------------------+
|peak_height        |number of reads at peak                                  |
+-------------------+---------------------------------------------------------+
|peak_median        |median coverage compared to peak height                  |
+-------------------+---------------------------------------------------------+
|interval_width     |width of interval                                        |
+-------------------+---------------------------------------------------------+
|peak_width         |width of peak                                            |
+-------------------+---------------------------------------------------------+
|bins               |bins for a histogram of densities within the interval.   |
+-------------------+---------------------------------------------------------+
|npeaks             |number of density peaks in interval.                     |
+-------------------+---------------------------------------------------------+
|peak_center        |point of highest density in interval                     |
+-------------------+---------------------------------------------------------+
|peak_relative_pos  |point of highest density in interval coordinates         |
+-------------------+---------------------------------------------------------+
|counts             |counts for a histogram of densities within the interval  |
+-------------------+---------------------------------------------------------+
|furthest_half_heigh|Distance of peak center to furthest half-height position |
+-------------------+---------------------------------------------------------+
|closest_half_height|Distance of peak center to closest half-height position  |
+-------------------+---------------------------------------------------------+


Additionally, the script outputs a set of matrixes with densities over
intervals that can be used for plotting. The default filenames are
``(matrix|control)_<sortorder>.tsv.gz``, The names can be controlled
with the ``--output-filename-pattern`` option.


Type::

   python bam2peakshape.py --help

for command line help.


Options
-------

Option: Shift
+++++++++++++

shift the each read by a certain distance, because in a ChIP-Seq
experment, the read is always at the edge of an sonicated fragment,
the actual binding site is usually L/2 distance away from the read,
where L is the length of sonicated fragment (determined either
experimentally or computationally).

This option is used only if the input reads are in :term:`bam` formatted file.
If input reads are :term:`bigwig` formatted file, this option is ignored.

Option: Random shift
++++++++++++++++++++

randomly shift the set of input regions to generate a artificial set
of regions. In the absence of real ChIP-Seq control library, the
random regions can provide a peaks profile that can be used as the
control.

Option: Centring method
+++++++++++++++++++++++

"reads" will output in the way that the summit of the peaks are
aligned. "middle" will output in the way that the middle of the input
bed intervals are aligned.

Option: Only interval
+++++++++++++++++++++

Only count reads that are in the interval as defined by the input bed file.

Option: normalization=sum
+++++++++++++++++++++++++

normalize counts such that the sum of all counts in all features are
exactly 1000000.

The detail normalization algorithm as follows: norm = sum(all counts
in all features)/1000000.0 normalized count = normalized count / norm

.. todo::

   paired-endedness is not fully implemented.

Command line options
--------------------

'''

import sys
import os
import re
import CGAT.Experiment as E
import CGAT.IOTools as IOTools
import pysam
import CGAT.Bed as Bed
import numpy
import collections
import pyBigWig

import CGAT.scripts._bam2peakshape as _bam2peakshape


def buildOptionParser(argv):

    if not argv:
        argv = sys.argv

    # setup command line parser
    parser = E.OptionParser(version="%prog version: $Id",
                            usage=globals()["__doc__"])

    parser.add_option("-f", "--format", dest="format", type="choice",
                      choices=("bam", "bigwig"),
                      help="format of genomic input files for densities "
                      "[%default]")

    parser.add_option(
        "-o", "--use-interval", dest="use_interval", action="store_true",
        help="only count tags that are in interval given "
        "in bed file. Otherwise, use a fixed width window (see --window-size) "
        "around peak [%default]")

    parser.add_option(
        "-w", "--window-size", dest="window_size", type="int",
        help="window size in bp on either side of a peak used for getting "
        "read densities. If ``--window-size`` is 1000, the actual window size"
        "will be 2kb, 1kb on either side of the peak in an interval"
        "[%default]")

    parser.add_option(
        "-b", "--bin-size", dest="bin_size", type="int",
        help="bin-size in bp for computing read densities. "
        "If ``--window-size`` is set to 1000 and ``--bin-size`` to 10, "
        "there will be 100 bins on either side of a peak. "
        "[%default]")

    parser.add_option(
        "--smooth-method", dest="smooth_method", type="choice",
        choices=("none", "sum", "sg"),
        help="smooting method to apply to density data before sampling "
        "according to ``bin-size``. sg=SavitzkyGolay, sum=sum density in bin, "
        "none=no smoothing "
        "[%default]")

    parser.add_option("-s", "--sort-order", dest="sort_orders",
                      type="choice",
                      action="append",
                      choices=("peak-height", "peak-width", "unsorted",
                               "interval-width", "interval-score"),
                      help="output sort order for matrices. "
                      "[%default]")

    parser.add_option(
        "-c", "--control-bam-file", "--control-bigwig-file",
        action="append",
        dest="control_files",
        type="string",
        help="control file. If given, two peakshapes are computed, "
        "one for the primary data and one for the control data. "
        "The control file is centered around the same "
        "base as the primary file and output in the same "
        "sort order as the primary profile to all side-by-side. "
        "comparisons. Multiple control files can be given. The "
        "control files should have the same format as the "
        "principal input file "
        "[%default]")

    parser.add_option(
        "-r", "--random-shift", dest="random_shift", action="store_true",
        help="shift intervals in random direction up/downstream of interval "
        "[%default]")

    parser.add_option(
        "-e", "--centring-method", dest="centring_method", type="choice",
        choices=("reads", "middle"),
        help="centring method. Available are: "
        "reads=use density to determine peak, "
        "middle=use middle of interval "
        "[%default]")

    parser.add_option(
        "-n", "--normalize-matrix", dest="normalization", type="choice",
        choices=("none", "sum"),
        help="matrix normalisation to perform. "
        "[%default]")

    parser.add_option(
        "--use-strand", dest="strand_specific", action="store_true",
        help="use strand information in intervals. Intervals on the "
        "negative strand are flipped "
        "[%default]")

    parser.add_option(
        "-i", "--shift-size", dest="shift", type="int",
        help="shift for reads. When processing bam files, "
        "reads will be shifted upstream/downstream by this amount. "
        "[%default]")

    parser.set_defaults(
        bin_size=10,
        shift=0,
        window_size=1000,
        sort_orders=[],
        centring_method="reads",
        control_files=[],
        random_shift=False,
        strand_specific=False,
        format="bam",
        report_step=100,
        use_interval=False,
        smooth_method=None,
    )

    return parser


IntervalData = collections.namedtuple(
    "IntervalData",
    "foreground interval controls shifted")


def outputFeatureTable(outfile, features_per_interval, bins):
    '''ouput results from density profiles.'''

    outfile.write("\t".join(
        ("contig",
         "start",
         "end",
         "name",
         "\t".join(_bam2peakshape.PeakShapeResult._fields))) + "\n")

    # output principal table
    n = 0
    for foreground, bed, controls, shifted in features_per_interval:
        n += 1
        if "name" in bed:
            name = bed.name
        else:
            name = str(n)
        outfile.write("%s\t%i\t%i\t%s\t" %
                      (bed.contig, bed.start, bed.end, name))

        outfile.write("\t".join(map(str, foreground[:-2])))
        bins, counts = foreground[-2], foreground[-1]
        outfile.write("\t%s" % ",".join(map(str, bins)))
        outfile.write("\t%s" % ",".join(map(str, counts)))
        outfile.write("\n")


def writeMatricesForSortOrder(features_per_interval,
                              bins,
                              foreground_track,
                              control_tracks,
                              shifted,
                              sort_order):
    '''output one or more matrices for each sort sorder.

    For each sort order output the forerground. If there
    are additional controls and shifted section, output
    these as well

    The files will named:
    matrix_<track>_<sortorder>

    '''
    if "name" in features_per_interval[0].interval:
        names = [x.interval.name for x in features_per_interval]
    else:
        names = list(map(str, list(range(1, len(features_per_interval) + 1))))

    bins = ["%i" % x for x in bins]
    sort_order = re.sub("-", "_", sort_order)

    # write foreground
    IOTools.writeMatrix(
        E.openOutputFile("matrix_%s_%s.gz" % (foreground_track, sort_order)),
        [x.foreground.counts for x in features_per_interval],
        row_headers=names,
        col_headers=bins,
        row_header="name")

    # write controls
    for idx, track in enumerate(control_tracks):
        IOTools.writeMatrix(
            E.openOutputFile("matrix_%s_%s.gz" % (track, sort_order)),
            [x.controls[idx].counts for x in features_per_interval],
            row_headers=names,
            col_headers=bins,
            row_header="name")

    # write shifted matrix
    if shifted:
        IOTools.writeMatrix(
            E.openOutputFile("matrix_shift_%s.gz" % (sort_order)),
            [x.shifted.counts for x in features_per_interval],
            row_headers=names,
            col_headers=bins,
            row_header="name")

    # output a combined matrix
    if len(control_tracks) > 0 or shifted:
        rows = []
        for row in features_per_interval:
            l = [row.foreground.counts]
            l.extend([row.controls[x].counts for x in
                      range(len(control_tracks))])
            if shifted:
                l.append(row.shifted.counts)
            rows.append(numpy.concatenate(l))

        n = 1 + len(control_tracks)
        if shifted:
            n += 1

        # make column names unique and make sure they can be sorted
        # lexicographically
        all_bins = []
        for x in range(n):
            all_bins.extend(["%i:%s" % (x, b) for b in bins])

        IOTools.writeMatrix(
            E.openOutputFile("matrix_sidebyside_%s.gz" % (sort_order)),
            rows,
            row_headers=names,
            col_headers=all_bins,
            row_header="name")


def outputMatrices(features_per_interval,
                   bins,
                   foreground_track,
                   control_tracks=None,
                   shifted=False,
                   sort_orders=None):
    '''ouput matrices from density profiles
    in one or more sort_orders.
    '''

    # output sorted matrices
    if not sort_orders:
        writeMatricesForSortOrder(features_per_interval,
                                  bins,
                                  foreground_track,
                                  control_tracks,
                                  shifted,
                                  "unsorted")

    for sort_order in sort_orders:

        if sort_order == "peak-height":
            features_per_interval.sort(
                key=lambda x: x.foreground.peak_height)

        elif sort_order == "peak-width":
            features_per_interval.sort(
                key=lambda x: x.foreground.peak_width)

        elif sort_order == "interval-width":
            features_per_interval.sort(
                key=lambda x: x.interval.end - x.interval.start)

        elif sort_order == "interval-score":
            try:
                features_per_interval.sort(
                    key=lambda x: float(x.interval.score))
            except IndexError:
                E.warn("score field not present - no output")
                continue
            except TypeError:
                E.warn("score field not a valid number - no output")
                continue

        writeMatricesForSortOrder(features_per_interval,
                                  bins,
                                  foreground_track,
                                  control_tracks,
                                  shifted,
                                  sort_order)


def buildDensityMatrices(bedfile,
                         fg_file,
                         control_files,
                         counter,
                         window_size=1000,
                         bin_size=10,
                         strand_specific=False,
                         centring_method="reads",
                         use_interval=False,
                         random_shift=False,
                         smooth_method="none",
                         report_step=1000):
    '''compute densities and peakshape parameters
    in intervals given by *bedfile* using reads in *fg_file*.

    If *control_files* are given, densities are produced for
    these as well.

    Returns a list of results for each interval in *bedfile* of
    type IntervalData and an array of bin-values.
    '''

    if window_size:
        # bins are centered at peak-center and then stretching outwards.
        bins = numpy.arange(-window_size + bin_size // 2,
                            +window_size,
                            bin_size)

    result = []
    c = E.Counter()
    c.input = 0

    for bed in bedfile:
        c.input += 1

        # if bed.contig not in contigs:
        #    c.skipped += 1
        #    continue

        if c.input % report_step == 0:
            E.info("iteration: %i" % c.input)

        features = counter.countInInterval(
            fg_file,
            bed.contig, bed.start, bed.end,
            window_size=window_size,
            bins=bins,
            use_interval=use_interval,
            centring_method=centring_method)

        if features is None:
            c.skipped += 1
            continue

        if control_files:
            control = []
            for control_file in control_files:
                control.append(counter.countAroundPos(
                    control_file,
                    bed.contig,
                    features.peak_center,
                    bins=features.bins))

        else:
            control = None

        if random_shift:
            direction = numpy.random.randint(0, 2)
            if direction:
                pos = features.peak_center + 2 * bins[0]
            else:
                pos = features.peak_center + 2 * bins[-1]
            shifted = counter.countAroundPos(fg_file,
                                             bed.contig,
                                             pos,
                                             bins=features.bins)
        else:
            shifted = None

        if strand_specific and bed.strand == "-":
            features._replace(hist=features.hist[::-1])
            if control:
                for c in control:
                    c._replace(hist=c.hist[::-1])
            if shifted:
                shifted._replace(hist=shifted.hist[::-1])

        result.append(IntervalData._make((features, bed, control, shifted)))
        c.added += 1

    E.info("interval processing: %s" % c)

    return result, bins


def main(argv=None):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    parser = buildOptionParser(argv)

    # add common options (-h/--help, ...) and parse command line
    (options, args) = E.Start(parser, argv=argv, add_output_options=True)

    if len(args) != 2:
        raise ValueError(
            "please specify one bam- or wig-file and one bed file")

    if options.control_files:
        E.info("using control files: %s" % ",".join(options.control_files))

    infile, bedfile = args
    control_files = []

    if options.format == "bigwig":
        fg_file = pyBigWig.open(infile)
        for control_file in options.control_files:
            control_files.append(pyBigWig.open(control_file))
        counter = _bam2peakshape.CounterBigwig(
            smooth_method=options.smooth_method)

    elif options.format == "bam":
        fg_file = pysam.AlignmentFile(infile, "rb")
        for control_file in options.control_files:
            control_files.append(pysam.AlignmentFile(control_file, "rb"))
        counter = _bam2peakshape.CounterBam(
            shift=options.shift,
            smooth_method=options.smooth_method)

    features_per_interval, bins = buildDensityMatrices(
        Bed.iterator(IOTools.openFile(bedfile)),
        fg_file,
        control_files,
        counter,
        window_size=options.window_size,
        bin_size=options.bin_size,
        strand_specific=options.strand_specific,
        centring_method=options.centring_method,
        use_interval=options.use_interval,
        random_shift=options.random_shift,
        smooth_method=options.smooth_method,
        report_step=options.report_step)

    if len(features_per_interval) == 0:
        E.warn("no data - no output")
        E.Stop()
        return

    outputFeatureTable(options.stdout, features_per_interval, bins)

    # apply normalization
    # Note: does not normalize control?
    # Needs reworking, currently it does not normalize across
    # all samples nor does the work "sum" reflect the per million
    # normalization.
    if options.normalization == "sum":
        E.info("starting sum normalization")
        # get total counts across all intervals
        norm = 0.0
        for foreground, bed, controls, shifted in features_per_interval:
            norm += sum(foreground.counts)
        # per million
        norm /= float(1000000)
        E.info("sum/million normalization with %f" % norm)

        # normalise
        new_data = []
        for foreground, bed, controls, shifted in features_per_interval:

            foreground = foreground._replace(
                counts=numpy.array(foreground.counts,
                                   dtype=numpy.float) / norm)
            new_controls = []
            for control in controls:
                new_controls.append(
                    control._replace(
                        counts=numpy.array(control.counts,
                                           dtype=numpy.float) / norm))
            if shifted:
                shifted = shifted._replace(
                    counts=numpy.array(shifted.counts,
                                       dtype=numpy.float) / norm)
            new_data.append(IntervalData._make((
                foreground, bed, new_controls, shifted)))
        features_per_interval = new_data
    else:
        E.info("no normalization performed")

    # center bins
    out_bins = bins[:-1] + options.bin_size

    # build tracks
    def _toTrack(filename):
        return os.path.splitext(os.path.basename(filename))[0]

    outputMatrices(features_per_interval,
                   out_bins,
                   foreground_track=_toTrack(infile),
                   control_tracks=[_toTrack(x) for x in options.control_files],
                   shifted=options.random_shift,
                   sort_orders=options.sort_orders)

    # write footer and output benchmark information.
    E.Stop()

if __name__ == "__main__":
    sys.exit(main(sys.argv))




'''
Masker.py - Wrapper for sequence masking tools
==============================================

:Tags: Python

Code
----

'''
import os
import subprocess
import tempfile
import string
import re
import random

from CGAT import Experiment as E
from CGAT import Genomics as Genomics
from CGAT import FastaIterator as FastaIterator
from six import StringIO


class Masker:

    """a masker preserves gaps, but it does not preserve
    whitespace characters."""

    mExecutable = None
    mOptions = ""
    mHasPeptideMasking = False
    mHasNucleicAcidMasking = False

    # set to true if masker outputs softmasked sequence
    soft_mask = False

    def __init__(self):
        pass

    def getAlphabet(self, sequence):
        """get sequence type (aa,na,codons)."""

        s1 = re.sub("[acgtxn\-.]", "", sequence.lower())
        s2 = re.sub("[-.]", "", sequence.lower())
        if float(len(s1)) < (len(s2) * 0.1):
            alphabet = "na"
            if len(sequence) % 3 == 0:
                alphabet = "codons"
        else:
            alphabet = "aa"

        return alphabet

    def __call__(self, sequence):
        """mask a sequence."""

        sequence = re.sub("\s", "", sequence)

        a = self.getAlphabet(sequence)

        seq = list(sequence)

        if len(seq) < 5:
            # do not mask empty/short sequences
            pass

        elif a == "aa" and self.mHasPeptideMasking:

            c = 0
            m = self.maskSequence(sequence)
            if self.soft_mask:
                m = re.sub("[a-z]", "x", m)
            for p, m in zip(sequence, m):
                if m in "Xx":
                    if p.isupper():
                        seq[c] = "X"
                    else:
                        seq[c] = "x"
                c += 1

        elif a == "codons" and self.mHasPeptideMasking:

            peptide_sequence = Genomics.TranslateDNA2Protein(sequence)
            masked_sequence = self.maskSequence(peptide_sequence)
            if self.soft_mask:
                masked_sequence = re.sub("[a-z]", "x", masked_sequence)

            c = 0
            for p, m in zip(peptide_sequence, masked_sequence):
                if m in "Xx":
                    if p.isupper():
                        seq[c:c + 3] = ["N"] * 3
                    else:
                        seq[c:c + 3] = ["n"] * 3
                c += 3

        elif a in ("na", "codons") and self.mHasNucleicAcidMasking:
            masked_sequence = self.maskSequence(sequence)
            if self.soft_mask:
                masked_sequence = re.sub("[a-z]", "N", masked_sequence)
            return masked_sequence
        else:
            raise ValueError(
                "masking of sequence type %s not implemented." % a)

        return "".join(seq)

    def maskSequence(self, peptide_sequence):
        """mask peptide sequence
        """

        Masker.__init__(self)

        outfile, filename_peptide = tempfile.mkstemp()
        os.write(outfile, ">test\n%s\n" % (peptide_sequence))
        os.close(outfile)

        infile = filename_peptide
        statement = self.mCommand % locals()

        E.debug("statement: %s" % statement)

        s = subprocess.Popen(statement,
                             shell=True,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE,
                             close_fds=True)

        (out, err) = s.communicate()
        if s.returncode != 0:
            raise RuntimeError(
                "Error in running %s \n%s\nTemporary directory" %
                (statement, err))

        os.remove(filename_peptide)

        masked_sequence = re.sub(
            "\s", "", string.join(out.split("\n")[1:], ""))
        return masked_sequence

    def maskSequences(self, sequences):
        '''mask a collection of sequences.'''

        with tempfile.NamedTemporaryFile(mode="w+t", delete=False) as outf:
            for x, s in enumerate(sequences):
                outf.write(">%i\n%s\n" % (x, s))

        infile = outf.name
        statement = self.mCommand % locals()

        E.debug("statement: %s" % statement)

        s = subprocess.Popen(statement,
                             shell=True,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE,
                             close_fds=True)

        (out, err) = s.communicate()

        if s.returncode != 0:
            raise RuntimeError(
                "Error in running %s \n%s\nTemporary directory" %
                (statement, err))
        
        result = [
            x.sequence for x in FastaIterator.iterate(StringIO(out.decode()))]

        os.remove(infile)

        return result


class MaskerBias (Masker):

    mCommand = "biasdb.pl %(infile)s"
    mHasPeptideMasking = True


class MaskerSeg (Masker):
    # mCommand = "seg %(infile)s 12 2.2 2.5 -x"
    mCommand = ("segmasker -in %(infile)s -window 12 -locut 2.2 "
                "-hicut 2.5 -outfmt fasta")
    mHasPeptideMasking = True
    soft_mask = True


class MaskerDustMasker(Masker):

    '''use dustmasker. masked chars are returned as
    lower case characters.'''

    mCommand = "dustmasker -outfmt fasta -in %(infile)s"
    mHasNucleicAcidMasking = True


class MaskerRandom (Masker):
    """randomly mask a proportion of positions in a sequence
    in multiple alignment."""

    def __init__(self, proportion=10, *args, **kwargs):
        Masker.__init__(self, *args, **kwargs)
        self.mProportion = proportion / 100.0

    def __call__(self, sequence):
        """mask a sequence."""

        sequence = re.sub("\s", "", sequence)

        a = self.getAlphabet(sequence)

        if a == "codons":
            frame = 3
        else:
            frame = 1

        positions = [
            x for x in range(0, len(sequence), frame) if sequence[x] != "-"]
        to_mask = random.sample(
            positions, int(len(positions) * self.mProportion))
        print(int(len(positions) * self.mProportion))

        s = list(sequence)
        print(to_mask)
        for x in to_mask:
            for y in range(x, x + frame):
                s[x] == "x"

        return "".join(s)

if __name__ == "__main__":
    x = MaskerRandom()
    print(x("AAA AAA AAA AAA --- AAA AAA AAA AAA"))

    x = MaskerDustMasker()
    print(x.maskSequences(("AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
                           "GGGGGGGGGG", )))


def maskSequences(sequences, masker=None):
    '''return a list of masked sequence.

    *masker* can be one of
        dust/dustmasker * run dustmasker on sequences
        softmask        * use softmask to hardmask sequences
    '''

    if masker in ("dust", "dustmasker"):
        masker_object = MaskerDustMasker()
    else:
        masker_object = None

    if masker == "softmask":
        # the genome sequence is repeat soft-masked
        masked_seq = sequences
    elif masker in ("dust", "dustmasker"):
        # run dust
        masked_seq = masker_object.maskSequences(
            [x.upper() for x in sequences])
    elif masker is None:
        masked_seq = [x.upper() for x in sequences]
    else:
        raise ValueError("unknown masker %s" % masker)

    # hard mask softmasked characters
    masked_seq = [re.sub("[a-z]", "N", x) for x in masked_seq]

    return masked_seq



"""beds2beds.py - decompose bed files
==================================

:Tags: Genomics Intervals BED Manipulation

Purpose
-------

This script will decompose a collection of input bedfiles into a
collection of unions or intersections.

Options
-------

Files are collected by a regular expression pattern given to the option
``--pattern-identifier``.

The script behaviour is determined by the ``--method`` option with either of
the following choices:

``merged-combinations``
    merge intervals across :term:`bed` files and only report those
    that appear in every file.

``unmerged-combinations``
    for each :term:`bed` file, report intervals that overlap with intervals
    in every other :term:`bed` file.

If the ``--exclusive-overlap`` option is set, report exclusive
overlap. Only intervals will be reported that overlap in a pairwise
comparison but do not overlap with intervals in any of the other sets.

This script requires bed files indexed by tabix_.

Usage
-----

For example, you have ChIP-Seq data for PolII and two transcription
factors tf1 and tf2. The following statement will output four
:term:`bed` files::

  zcat polii.bed.gz | head

  chr17    1    100    8    1
  chr19   -50    50    6    1
  chr19    0    100    1    1
  chr19    50   150    1    1
  chr19   150   200    2    1
  chr19   201   300    3    1

  python beds2beds.py polii.bed.gz tf1.bed.gz tf2.bed.gz

  zcat tf1.bed.gz | head

  chr1    35736     40736    ENST000004173240    -
  chr1    60881     65881    ENST000005349900    +
  chr1    64090     69090    ENST000003351370    +
  chr1    362658    367658   ENST000004264060    +
  chr1    622034    627034   ENST000003328310    -
  chr1    716405    721405   ENST000003585330    +


The four files contain intervals, that

1. have PolII and tf1 present,
2. have PolII and tf2 present,
3. have tf1 and tf2 present, or
4. have PolII and tf1 and tf2 present.

If the --exclusive-overlap option is set, three sets will be output
with intervals that

1. have PolII and tf1 present but no tf2,
2. have PolII and tf2 present but no tf1,
3. have tf1 and tf2 present but no PolII.

Type::

   python beds2beds.py --help

for command line help.

Command line options
--------------------

"""

import sys
import re
import itertools
import collections

import CGAT.Experiment as E
import CGAT.IOTools as IOTools
import pysam
import CGAT.Intervals as Intervals


def isContainedInAll(contig, start, end, bedfiles):

    for bedfile in bedfiles:
        try:
            if len(list(bedfile.fetch(contig, start, end))) == 0:
                return False
        except KeyError:
            return False
        except ValueError:
            return False

    return True


def isContainedInOne(contig, start, end, bedfiles):

    for bedfile in bedfiles:
        try:
            if len(list(bedfile.fetch(contig, start, end))) > 0:
                return True
        except KeyError:
            pass
        except ValueError:
            pass

    return False


def combineMergedIntervals(bedfiles):
    '''combine intervals in a collection of bed files.

    Overlapping intervals between tracks are merged.

    Algorithm:

    1. collect all intervals in all tracks into a single track
    2. merge overlapping intervals
    3. report all intervals that overlap with an interval in each track.

    '''

    # get all intervals
    data_per_contig = collections.defaultdict(list)

    for bedfile in bedfiles:
        for contig in bedfile.contigs:
            i = []
            for bed in bedfile.fetch(contig, parser=pysam.asBed()):
                i.append((bed.start, bed.end))
            data_per_contig[contig].extend(i)

    # merge intervals
    for contig in list(data_per_contig.keys()):
        data_per_contig[contig] = Intervals.combine(data_per_contig[contig])

    # filter intervals - take only those present in all bedfiles
    for contig, data in sorted(data_per_contig.items()):
        for start, end in data:
            if isContainedInAll(contig, start, end, bedfiles):
                yield contig, start, end


def combineUnmergedIntervals(foreground, background):
    '''combine intervals in a collection of bed files.

    Only intervals in the first track are reported.

    Algorithm:

    1. report all intervals in the first track that overlap with an
    interval in every other track.

    '''

    c = 0
    for bed in foreground.fetch(parser=pysam.asBed()):
        c += 1
        if isContainedInAll(bed.contig, bed.start, bed.end, background):
            yield bed


def main(argv=None):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv:
        argv = sys.argv

    # setup command line parser
    parser = E.OptionParser(
        version="%prog version: $Id$",
        usage=globals()["__doc__"])

    parser.add_option(
        "-e", "--exclusive-overlap", dest="exclusive",
        action="store_true",
        help="Intervals reported will be merged across the "
        "positive set and do not overlap any interval in any of the "
        "other sets [default=%default].")

    parser.add_option(
        "-p", "--pattern-identifier", dest="pattern_id", type="string",
        help="pattern to convert a filename "
        "to an id [default=%default].")

    parser.add_option(
        "-m", "--method", dest="method", type="choice",
        choices=("merged-combinations",
                 "unmerged-combinations"),
        help="method to perform [default=%default]")

    parser.set_defaults(
        pattern_id="(.*).bed.gz",
        exclusive=False,
        method="merged-combinations",
    )

    # add common options (-h/--help, ...) and parse command line
    (options, args) = E.Start(parser, argv=argv, add_output_options=True)

    if len(args) < 2:
        raise ValueError("at least two arguments required")

    tags, bedfiles = [], []
    for infile in args:
        bedfiles.append(pysam.Tabixfile(infile, "r"))
        tags.append(re.search(options.pattern_id, infile).groups()[0])

    indices = list(range(len(bedfiles)))
    is_exclusive = options.exclusive

    if options.method == "merged-combinations":

        if is_exclusive:
            start = 1
        else:
            start = 2

        options.stdout.write("combination\twithout\tcounts\n")

        for ncombinants in range(start, len(bedfiles) + 1):
            for combination in itertools.combinations(indices, ncombinants):
                other = [x for x in indices if x not in combination]
                tag = ":".join([tags[x] for x in combination])
                E.debug("combination %s started" % tag)
                E.debug("other: %s" % ":".join([tags[x] for x in other]))

                other_bed = [bedfiles[x] for x in other]
                outf = IOTools.openFile(
                    E.getOutputFile(tag), "w", create_dir=True)
                c = E.Counter()
                for contig, start, end in combineMergedIntervals(
                        [bedfiles[x] for x in combination]):
                    c.found += 1
                    if is_exclusive and isContainedInOne(contig,
                                                         start,
                                                         end,
                                                         other_bed):
                        c.removed += 1
                        continue
                    c.output += 1
                    outf.write("%s\t%i\t%i\n" % (contig, start, end))

                outf.close()
                E.info("combination %s finished: %s" % (tag, c))

                options.stdout.write("%s\t%s\t%i\n" % (
                    ":".join([tags[x] for x in combination]),
                    ":".join([tags[x] for x in other]),
                    c.output))

    elif options.method == "unmerged-combinations":
        options.stdout.write("track\tcombination\twithout\tcounts\n")

        for foreground in indices:

            start = 0

            background = [x for x in indices if x != foreground]
            for ncombinants in range(0, len(background) + 1):
                for combination in itertools.combinations(background,
                                                          ncombinants):
                    other = [x for x in background if x not in combination]
                    combination_bed = [bedfiles[x] for x in combination]
                    other_bed = [bedfiles[x] for x in other]
                    tag = ":".join([tags[foreground]] + [tags[x]
                                                         for x in combination])

                    E.debug("fg=%i, combination=%s, other=%s" %
                            (foreground, combination, other))
                    E.debug("combination %s started" % tag)
                    E.debug("other: %s" % ":".join([tags[x] for x in other]))

                    outf = IOTools.openFile(
                        E.getOutputFile(tag), "w", create_dir=True)
                    c = E.Counter()
                    for bed in combineUnmergedIntervals(
                            bedfiles[foreground],
                            combination_bed):
                        c.found += 1
                        if is_exclusive and isContainedInOne(bed.contig,
                                                             bed.start,
                                                             bed.end,
                                                             other_bed):
                            c.removed += 1
                            continue
                        c.output += 1
                        outf.write("%s\n" % str(bed))

                    outf.close()
                    E.info("combination %s finished: %s" % (tag, c))

                    options.stdout.write("%s\t%s\t%s\t%i\n" % (
                        tags[foreground],
                        ":".join([tags[x] for x in combination]),
                        ":".join([tags[x] for x in other]),
                        c.output))

    E.Stop()


if __name__ == "__main__":
    sys.exit(main(sys.argv))



'''gtfs2tsv.py - compare two genesets
==================================

:Tags: Python

Purpose
-------

This script compares two genesets (required) in :term:`gtf`-formatted
files and output lists of shared and unique genes.

It outputs the results of the comparison into various sections. The
sections are split into separate output files whose names are
determined by the ``--output-filename-pattern`` option. The sections
are:

``genes_ovl``
   Table with overlapping genes

``genes_total``
   Summary statistic of overlapping genes

``genes_uniq1``
   List of genes unique in set 1

``genes_uniq2``
   List of genes unique in set 2

Options
-------

``--output-filename-pattern``
   This option defines how the output filenames are determined for the
   sections described in the :term:`Purpose` section above.


Usage
-----

Example::

   head a.gtf::

     19 processed_transcript exon 66346 66509 . - . gene_id "ENSG00000225373";
     transcript_id "ENST00000592209"; exon_number "1"; gene_name "AC008993.5";
     gene_biotype "pseudogene"; transcript_name "AC008993.5-002";
     exon_id "ENSE00001701708";

     19 processed_transcript exon 60521 60747 . - . gene_id "ENSG00000225373";
     transcript_id "ENST00000592209"; exon_number "2"; gene_name "AC008993.5";
     gene_biotype "pseudogene"; transcript_name "AC008993.5-002";
     exon_id "ENSE00002735807";

     19 processed_transcript exon 60105 60162 . - . gene_id "ENSG00000225373";
     transcript_id "ENST00000592209"; exon_number "3"; gene_name "AC008993.5";
     gene_biotype "pseudogene"; transcript_name "AC008993.5-002";
     exon_id "ENSE00002846866";

   head b.gtf::

     19 transcribed_processed_pseudogene exon 66320 66492 . - .
     gene_id "ENSG00000225373"; transcript_id "ENST00000587045"; exon_number "1";
     gene_name "AC008993.5"; gene_biotype "pseudogene";
     transcript_name "AC008993.5-001"; exon_id "ENSE00002739353";

     19 lincRNA exon 68403 69146 . + . gene_id "ENSG00000267111";
     transcript_id "ENST00000589495"; exon_number "1"; gene_name "AC008993.2";
     gene_biotype "lincRNA"; transcript_name "AC008993.2-001";
     exon_id "ENSE00002777656";

     19 lincRNA exon 71161 71646 . + . gene_id "ENSG00000267588";
     transcript_id "ENST00000590978"; exon_number "1"; gene_name "MIR1302-2";
     gene_biotype "lincRNA"; transcript_name "MIR1302-2-001";
     exon_id "ENSE00002870487";

   python gtfs2tsv.py a.gtf b.gtf > out.tsv

   head out.tsv::

     contigs source feature start end score strand frame gene_id transcript_id attributes
     19 processed_transcript exon 66345 66509 . - . ENSG00000225373 ENST00000592209 exon_number "1";
     gene_name "AC008993.5"; gene_biotype "pseudogene"; transcript_name "AC008993.5-002";
     exon_id "ENSE00001701708"
     19 processed_transcript exon 60520 60747 . - . ENSG00000225373 ENST00000592209 exon_number "2";
     gene_name "AC008993.5"; gene_biotype "pseudogene"; transcript_name "AC008993.5-002";
     exon_id "ENSE00002735807"

Type::

   python gtfs2tsv.py --help

for command line help.

Command line options
--------------------

'''
import sys
import os
import CGAT.Experiment as E
import CGAT.GTF as GTF
import CGAT.IOTools as IOTools
import bx.intervals.intersection


def GetNextLine(infile):
    for line in infile:
        if line[0] == "#":
            continue
        return line
    return None


class Counts:

    mPercentFormat = "%.2f"

    def __init__(self, add_percent):

        self.nleft, self.nright, self.noverlap = 0, 0, 0
        self.nunique_left = 0
        self.nunique_right = 0
        self.nidentical = 0
        self.nhalf = 0
        self.nsplit_left = 0
        self.nsplit_right = 0
        self.mAddPercent = add_percent

    def __add__(self, other):
        self.nleft += other.nleft
        self.nright += other.nright
        self.noverlap += other.noverlap
        self.nunique_left += other.nunique_left
        self.nunique_right += other.nunique_right
        self.nidentical += other.nidentical
        self.nhalf += other.nhalf
        self.nsplit_left += other.nsplit_left
        self.nsplit_right += other.nsplit_right
        return self

    def getHeader(self):
        h = "total_left\ttotal_right\tnoverlap\tnidentical\tnhalf\tunique_left\tunique_right\tsplit_left\tsplit_right"
        if self.mAddPercent:
            h += "\t" + self.getHeaderPercent()
        return h

    def __str__(self):
        h = "\t".join(map(str,
                          (self.nleft, self.nright,
                              self.noverlap, self.nidentical, self.nhalf,
                              self.nunique_left, self.nunique_right,
                              self.nsplit_left, self.nsplit_right)))
        if self.mAddPercent:
            h += "\t" + self.asPercent()

        return h

    def getHeaderPercent(self):
        return "\t".join(["pl%s\tpr%s" % (x, x) for x in ("overlap", "identical", "half", "unique", "split")])

    def asPercent(self):
        return "\t".join([self.mPercentFormat % (100.0 * x) for x in (
            float(self.noverlap) / self.nleft,
            float(self.noverlap) / self.nright,
            float(self.nidentical) / self.nleft,
            float(self.nidentical) / self.nright,
            float(self.nhalf) / self.nleft,
            float(self.nhalf) / self.nright,
            float(self.nunique_left) / self.nleft,
            float(self.nunique_right) / self.nright,
            float(self.nsplit_left) / self.nleft,
            float(self.nsplit_right) / self.nright)])


def getFile(options, section):

    if options.output_filename_pattern:
        outfile = IOTools.openFile(
            options.output_filename_pattern % section, "w")
        E.info("output for section '%s' goes to file %s" %
               (section, options.output_filename_pattern % section))
    else:
        outfile = options.stdout
        outfile.write("## section: %s\n" % section)
    return outfile


def writeDiff(outfile, symbol, genes):

    for gene in sorted(genes):
        for exon in sorted(gene):
            outfile.write("%s\t%s\n" % (symbol, str(exon)))


def main(argv=None):

    if not argv:
        argv = sys.argv

    parser = E.OptionParser(
        version="%prog version: $Id$",
        usage=globals()["__doc__"])

    parser.add_option(
        "-e", "--output-equivalent", dest="write_equivalent",
        action="store_true",
        help="write equivalent entries [default=%default].")

    parser.add_option(
        "-f", "--output-full", dest="write_full",
        action="store_true",
        help="write full gff entries [default=%default].")

    parser.add_option("-p", "--add-percent", dest="add_percent",
                      action="store_true",
                      help="add percentage columns [default=%default].")

    parser.add_option("-s", "--ignore-strand", dest="ignore_strand",
                      action="store_true",
                      help="ignore strand information [default=%default].")

    parser.set_defaults(
        write_equivalent=False,
        write_full=False,
        add_percent=False,
        ignore_strand=False,
        as_gtf=False,
    )

    (options, args) = E.Start(parser, argv, add_output_options=True)

    if len(args) != 2:
        raise ValueError("two arguments required")

    input_filename1, input_filename2 = args

    # duplicated features cause a problem. Make sure
    # features are non-overlapping by running
    # gff_combine.py on GFF files first.

    E.info("reading data started")

    idx, genes2 = {}, set()
    for e in GTF.readFromFile(IOTools.openFile(input_filename2, "r")):
        genes2.add(e.gene_id)
        if e.contig not in idx:
            idx[e.contig] = bx.intervals.intersection.Intersecter()
        idx[e.contig].add_interval(
            bx.intervals.Interval(e.start, e.end, value=e))

    overlaps_genes = []

    E.info("reading data finished: %i contigs" % len(idx))

    # outfile_diff and outfile_overlap not implemented
    # outfile_diff = getFile( options, "diff" )
    # outfile_overlap = getFile( options, "overlap" )
    overlapping_genes = set()

    genes1 = set()

    # iterate over exons
    with IOTools.openFile(input_filename1, "r") as infile:
        for this in GTF.iterator(infile):

            genes1.add(this.gene_id)

            try:
                intervals = idx[this.contig].find(this.start, this.end)
            except KeyError:
                continue

            others = [x.value for x in intervals]
            for other in others:
                overlapping_genes.add((this.gene_id, other.gene_id))

            # check for identical/half-identical matches
            output = None
            for other in others:
                if this.start == other.start and this.end == other.end:
                    output, symbol = other, "="
                    break
            else:
                for other in others:
                    if this.start == other.start or this.end == other.end:
                        output, symbol = other, "|"
                        break
                else:
                    symbol = "~"

    # if outfile_diff != options.stdout: outfile_diff.close()
    # if outfile_overlap != options.stdout: outfile_overlap.close()

    outfile = None
    ##################################################################
    ##################################################################
    ##################################################################
    # print gene based information
    ##################################################################
    if overlapping_genes:
        outfile = getFile(options, "genes_ovl")
        outfile.write("gene_id1\tgene_id2\n")
        for a, b in sorted(overlapping_genes):
            outfile.write("%s\t%s\n" % (a, b))
        if outfile != options.stdout:
            outfile.close()

        outfile_total = getFile(options, "genes_total")
        outfile_total.write(
            "set\tngenes\tnoverlapping\tpoverlapping\tnunique\tpunique\n")

        outfile = getFile(options, "genes_uniq1")
        b = set([x[0] for x in overlapping_genes])
        d = genes1.difference(b)
        outfile.write("gene_id1\n")
        outfile.write("\n".join(sorted(d)) + "\n")
        if outfile != options.stdout:
            outfile.close()
        outfile_total.write("%s\t%i\t%i\t%5.2f\t%i\t%5.2f\n" % (
            os.path.basename(input_filename1), len(
                genes1), len(b), 100.0 * len(b) / len(a),
            len(d), 100.0 * len(d) / len(genes1)))

        outfile = getFile(options, "genes_uniq2")
        b = set([x[1] for x in overlapping_genes])
        d = genes2.difference(b)
        outfile.write("gene_id2\n")
        outfile.write("\n".join(sorted(d)) + "\n")
        if outfile != options.stdout:
            outfile.close()

        outfile_total.write("%s\t%i\t%i\t%5.2f\t%i\t%5.2f\n" % (
            os.path.basename(input_filename2), len(
                genes2), len(b), 100.0 * len(b) / len(a),
            len(d), 100.0 * len(d) / len(genes2)))
        if outfile_total != options.stdout:
            outfile_total.close()

    E.Stop()

if __name__ == "__main__":
    sys.exit(main())



'''
Maq.py - Working with Maq formatted files
=========================================

This module provides a parser for the output of the maq mapper.
As this mapper is out of use, the contents here are of historical
interest only.

Reference
---------

'''


class Error(Exception):

    """Base class for exceptions in this module."""

    def __str__(self):
        return str(self.message)

    def _get_message(self, message):
        return self._message

    def _set_message(self, message):
        self._message = message
    message = property(_get_message, _set_message)


class ParsingError(Error):

    """Exception raised for errors while parsing

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message, line=None):
        if line:
            self.message = message + " at line:" + line
        else:
            self.message = message


class Match:

    """a maq match.
    """

    def __init__(self):
        self.mName = None
        self.contig = None
        self.start = 0
        self.strand = "+"
        self.mPairInsert = 0
        self.mPairFlag = 0
        self.mQuality = 0
        self.mQualitySingle = 0
        self.mQualityAlternative = 0
        self.mNMismatches = 0
        self.mMismatchQuality = 0
        self.mNMismatch0 = 0
        self.mNMismatch1 = 0
        self.mLength = 0

    def __str__(self):
        return "\t".join(map(str, (self.mName,
                                   self.contig,
                                   self.start + 1,
                                   self.strand,
                                   self.mPairInsert,
                                   self.mPairFlag,
                                   self.mQuality,
                                   self.mQualitySingle,
                                   self.mQualityAlternative,
                                   self.mNMismatches,
                                   self.mMismatchQuality,
                                   self.mNMismatch0,
                                   self.mNMismatch1,
                                   self.mLength)))

    def fromTable(self, data):

        if len(data) == 14:
            (self.mName,
             self.contig,
             self.start,
             self.strand,
             self.mPairInsert,
             self.mPairFlag,
             self.mQuality,
             self.mQualitySingle,
             self.mQualityAlternative,
             self.mNMismatches,
             self.mMismatchQuality,
             self.mNMismatches0,
             self.mNMismatches1,
             self.mLength) = data

        (self.start,
         self.mPairInsert,
         self.mPairFlag,
         self.mQuality,
         self.mQualitySingle,
         self.mQualityAlternative,
         self.mNMismatches,
         self.mMismatchQuality,
         self.mNMismatches0,
         self.mNMismatches1,
         self.mLength) = list(map(int, (
             self.start,
             self.mPairInsert,
             self.mPairFlag,
             self.mQuality,
             self.mQualitySingle,
             self.mQualityAlternative,
             self.mNMismatches,
             self.mMismatchQuality,
             self.mNMismatches0,
             self.mNMismatches1,
             self.mLength)))
        # maq starts counting at 1
        self.start -= 1


def iterator(infile):
    """iterate over the contents of a maq file.
    """
    while 1:
        line = infile.readline()
        if not line:
            raise StopIteration
        if line[0] == "#":
            continue
        if line.startswith("read\tcontig"):
            continue
        match = Match()
        match.fromTable(line[:-1].split())
        yield match



'''bed2stats.py - summary of bed file contents
============================================

:Tags: Genomics Intervals Summary BED

Purpose
-------

This script takes a :term:`bed`-formatted file as input and outputs the number
of intervals and bases in the bed file. Counts can be subdivided by setting
the ``--aggregate-by`` command line option:

``contig``
   output counts per contig (column 1)

``name``
   output counts grouped by the name field in the :term:`bed` formatted
   file (column 4)

``track``
   output counts per track in the :term:`bed` formatted file.

Note that a count of bases usually makes only sense if the intervals
submitted are non-overlapping.

If the option --add-percent is given, an additional column will output
the percent of the genome covered by intervals. This requires a
--genome-file to be given as well.

Usage
-----

To count the number of intervals, type::

   cgat bed2table < in.bed

+-----+--------+----------+------+
|track|ncontigs|nintervals|nbases|
+-----+--------+----------+------+
|all  |23      |556       |27800 |
+-----+--------+----------+------+

To count per contig::

   cgat bed2table --aggregate=contig < in.bed

+-----+--------+----------+------+
|track|ncontigs|nintervals|nbases|
+-----+--------+----------+------+
|chrX |1       |11        |550   |
+-----+--------+----------+------+
|chr13|1       |12        |600   |
+-----+--------+----------+------+
|chr12|1       |37        |1850  |
+-----+--------+----------+------+
|...  |...     |...       |...   |
+-----+--------+----------+------+

Type::

   cgat bed2table --help

for command line help.

Command line options
--------------------

'''
import sys
import collections
import CGAT.Bed as Bed
import CGAT.Experiment as E
import CGAT.IndexedFasta as IndexedFasta


class Counter:

    headers = ["ncontigs", "nintervals", "nbases"]
    headers_percent = ["ncontigs", "nintervals", "nbases", "pbases"]

    def __init__(self):
        self.intervals_per_contig = collections.defaultdict(int)
        self.bases_per_contig = collections.defaultdict(int)
        self.size = None

    def setSize(self, size):
        self.size = size

    def add(self, bed):
        self.intervals_per_contig[bed.contig] += 1
        self.bases_per_contig[bed.contig] += bed.end - bed.start

    def __str__(self):
        bases = sum(self.bases_per_contig.values())
        if self.size is None:
            return "%i\t%i\t%i" % (len(self.intervals_per_contig),
                                   sum(self.intervals_per_contig.values()),
                                   bases,
                                   )
        else:
            return "%i\t%i\t%i\t%5.2f" % (len(self.intervals_per_contig),
                                          sum(self.intervals_per_contig.values(
                                          )),
                                          sum(self.bases_per_contig.values()),
                                          100.0 * bases / self.size
                                          )


def main(argv=None):

    if argv is None:
        argv = sys.argv

    parser = E.OptionParser(
        version="%prog version: $Id$",
        usage=globals()["__doc__"])

    parser.add_option(
        "-g", "--genome-file", dest="genome_file", type="string",
        help="filename with genome [default=%default].")

    parser.add_option(
        "-a", "--aggregate-by", dest="aggregate", type="choice",
        choices=("name", "contig", "track", "none"),
        help="aggregate counts by feature [default=%default].")

    parser.add_option(
        "-p", "--add-percent", dest="add_percent", action="store_true",
        help="add percentages [default=%default].")

    parser.set_defaults(
        genome_file=None,
        aggregate="none",
        add_percent=False,
    )

    (options, args) = E.Start(parser, argv)

    # get files
    if options.genome_file:
        fasta = IndexedFasta.IndexedFasta(options.genome_file)
    else:
        if options.add_percent:
            raise ValueError("--add-percent option requires --genome-file")
        fasta = None

    if options.add_percent and not options.aggregate == "contig":
        raise NotImplementedError(
            "--add-percent option requires --aggregate=contig")

    counts = collections.defaultdict(Counter)
    total = Counter()
    output_totals = True

    if options.aggregate == "track":
        keyf = lambda x: x.track
    elif options.aggregate == "name":
        keyf = lambda x: x.name
    elif options.aggregate == "contig":
        keyf = lambda x: x.contig
    else:
        keyf = lambda x: "all"
        output_totals = False

    for bed in Bed.iterator(options.stdin):
        counts[keyf(bed)].add(bed)
        total.add(bed)

    outf = options.stdout

    key = "track"
    if options.add_percent:
        outf.write("%s\t%s\n" % (key, "\t".join(Counter.headers_percent)))
    else:
        outf.write("%s\t%s\n" % (key, "\t".join(Counter.headers)))

    total_bases = 0
    for key, count in sorted(counts.items()):
        if options.add_percent:
            total_bases += fasta.getLength(key)
            count.setSize(fasta.getLength(key))

        outf.write("%s\t%s\n" % (key, str(count)))

    if output_totals:
        if options.add_percent:
            count.setSize(total_bases)
        outf.write("%s\t%s\n" % ("total", str(total)))
    E.Stop()

if __name__ == '__main__':
    sys.exit(main(sys.argv))
